"""SC Controller - OSD Launcher

Display launcher with phone-like keyboard that user can use to select
application (list is generated using xdg) and start it.

Reuses styles from OSD Menu and OSD Dialog
"""

import base64
import logging
import os
import sys
from enum import IntEnum
from typing import Self

from gi.repository import Gtk

from scc.actions import Action, AxisAction, DPadAction, MouseAction, MultiAction, XYAction
from scc.config import Config
from scc.constants import SCButtons
from scc.gui.daemon_manager import DaemonManager
from scc.gui.svg_widget import SVGEditor, SVGWidget
from scc.modifiers import DoubleclickModifier, ModeModifier
from scc.osd import OSDWindow
from scc.parser import TalkingActionParser
from scc.paths import get_config_path, get_share_path
from scc.profile import Profile
from scc.special_actions import MenuAction
from scc.tools import _, nameof
from scc.uinput import Rels

log = logging.getLogger("osd.binds")


class BindingDisplay(OSDWindow):
	def __init__(self, config=None):
		self.bdisplay = os.path.join(get_config_path(), "binding-display.svg")
		if not os.path.exists(self.bdisplay):
			# Prefer image in ~/.config/scc, but load default one as fallback
			self.bdisplay = os.path.join(get_share_path(), "images", "binding-display.svg")

		OSDWindow.__init__(self, "osd-keyboard")
		self.daemon = None
		self.config = config or Config()
		self.group = None
		self.limits = {}
		self.background = None

		self._eh_ids = []
		self._stick = 0, 0

		self.c = Gtk.Box()
		self.c.set_name("osd-keyboard-container")

	def on_profile_changed(self, daemon: DaemonManager, filename: str):
		profile = Profile(TalkingActionParser()).load(filename)
		Generator(SVGEditor(self.background), profile)

	def use_daemon(self, d):
		"""Allows (re)using already existing DaemonManager instance in same process."""
		self.daemon = d
		self._cononect_handlers()
		self.on_daemon_connected(self.daemon)

	def _add_arguments(self):
		OSDWindow._add_arguments(self)
		self.argparser.add_argument("image", type=str, nargs="?", default=self.bdisplay, help="keyboard image to use")
		self.argparser.add_argument(
			"--cancel-with", type=str, metavar="button", default="B", help="button used to close display (default: B)",
		)

	def compute_position(self):
		"""Unlike other OSD windows, this one is scaled to 80% of screen size and centered in on active screen."""
		x, y = 10, 10
		iw, ih = self.background.image_width, self.background.image_height
		geometry = self.get_active_screen_geometry()
		if geometry:
			width, height = iw, ih
			if width > geometry.width * 0.8:
				width = geometry.width * 0.8
				height = int(float(ih) / float(iw) * float(width))
				self.background.resize(width, height)
				self.background.hilight({})
			x = geometry.x + ((geometry.width - width) / 2)
			y = geometry.y + ((geometry.height - height) / 2)
		return x, y

	def parse_argumets(self, argv):
		if not OSDWindow.parse_argumets(self, argv):
			return False
		self._cancel_with = self.args.cancel_with
		return True

	def _cononect_handlers(self):
		self._eh_ids += [
			(self.daemon, self.daemon.connect("dead", self.on_daemon_died)),
			(self.daemon, self.daemon.connect("error", self.on_daemon_died)),
			(self.daemon, self.daemon.connect("profile-changed", self.on_profile_changed)),
			(self.daemon, self.daemon.connect("alive", self.on_daemon_connected)),
		]

	def run(self):
		self.daemon = DaemonManager()
		self._cononect_handlers()
		OSDWindow.run(self)

	def on_daemon_connected(self, *a):
		def success(*a):
			log.info("Sucessfully locked input")

		c = self.choose_controller(self.daemon)
		if c is None or not c.is_connected():
			# There is no controller connected to daemon
			self.on_failed_to_lock("Controller not connected")
			return

		self._eh_ids += [
			(c, c.connect("event", self.on_event)),
			(c, c.connect("lost", self.on_controller_lost)),
		]

		# Lock everything
		locks = ["RB", "LB", self.args.cancel_with]
		c.lock(success, self.on_failed_to_lock, *locks)

	def quit(self, code=-1):
		if self.get_controller():
			self.get_controller().unlock_all()
		for source, eid in self._eh_ids:
			source.disconnect(eid)
		self._eh_ids = []
		OSDWindow.quit(self, code)

	def show(self, *a):
		if self.background is None:
			self.realize()
			self.background = SVGWidget(self.args.image, init_hilighted=True)
			self.c.add(self.background)
			self.add(self.c)

		OSDWindow.show(self, *a)
		self.move(*self.compute_position())

	def on_event(self, daemon, what, data):
		"""Called when button press, button release or stick / pad update is
		send by daemon.
		"""
		if what == self._cancel_with:
			if data[0] == 0:  # Button released
				self.quit(-1)


class Align(IntEnum):
	TOP = 1 << 0
	BOTTOM = 1 << 1
	LEFT = 1 << 2
	RIGHT = 1 << 3


def find_image(name):
	# TODO: This
	filename = "images/" + name + ".svg"
	if os.path.exists(filename):
		return filename
	return None


class Line:
	def __init__(self, icon, text):
		self.icons = [icon]
		self.text = text

	def get_size(self, gen):
		# TODO: This
		return gen.char_width * len(self.text), gen.line_height

	def add_icon(self, icon) -> Self:
		self.icons.append(icon)
		return self

	def to_string(self):
		return "%-10s: %s" % (",".join([x for x in self.icons if x]), self.text)


class LineCollection:
	"""Allows calling add_icon on multiple lines at once"""

	def __init__(self, *lines):
		self.lines = lines

	def add_icon(self, icon) -> Self:
		for line in self.lines:
			line.add_icon(icon)
		return self


class Box:
	PADDING = 5
	SPACING = 2
	MIN_WIDTH = 100
	MIN_HEIGHT = 50

	def __init__(self, anchor_x, anchor_y, align, name, min_width=MIN_WIDTH, min_height=MIN_HEIGHT, max_width=999999):
		self.name = name
		self.lines = []
		self.anchor = anchor_x, anchor_y
		self.align = align
		self.min_height = min_height
		self.x, self.y = 0, 0
		self.min_width = min_width
		self.max_width = max_width
		self.min_height = min_height

	def to_string(self):
		return "--- %s ---\n%s\n" % (self.name, "\n".join([x.to_string() for x in self.lines]))

	def add(self, icon, context, action):
		if not action:
			return LineCollection()
		if isinstance(action, MultiAction):
			if not action.is_key_combination():
				return LineCollection([self.add(icon, context, child) for child in action.actions])
		elif isinstance(action, ModeModifier):
			lines = [self.add(icon, context, action.default)]
			for x in action.mods:
				lines.append(self.add(nameof(x), context, action.mods[x]).add_icon(icon))
			return LineCollection(*lines)
		elif isinstance(action, DoubleclickModifier):
			lines = []
			if action.normalaction:
				lines.append(self.add(icon, context, action.normalaction))
			if action.action:
				lines.append(self.add("DOUBLECLICK", context, action.action).add_icon(icon))
			if action.holdaction:
				lines.append(self.add("HOLD", context, action.holdaction).add_icon(icon))
			return LineCollection(*lines)

		action = action.strip()
		if isinstance(action, MenuAction):
			if self.name == "bcs" and action.menu_id == "Default.menu":
				# Special case, this action is expected in every profile,
				# so there is no need to draw it here
				return LineCollection()
		elif isinstance(action, DPadAction):
			return LineCollection(
				self.add("DPAD_UP", Action.AC_BUTTON, action.actions[0]),
				self.add("DPAD_DOWN", Action.AC_BUTTON, action.actions[1]),
				self.add("DPAD_LEFT", Action.AC_BUTTON, action.actions[2]),
				self.add("DPAD_RIGHT", Action.AC_BUTTON, action.actions[3]),
			)
		elif isinstance(action, XYAction):
			if isinstance(action.x, MouseAction) and isinstance(action.y, MouseAction):
				if action.x.get_axis() in (Rels.REL_HWHEEL, Rels.REL_WHEEL):
					# Special case, pad bound to wheel
					line = Line(icon, _("Mouse Wheel"))
					self.lines.append(line)
					return line
			if isinstance(action.x, AxisAction) and isinstance(action.y, AxisAction):
				if action.x.axis and action.y.axis:
					line = Line(icon, action.x.describe(Action.AC_BUTTON))
					self.lines.append(line)
					return line
			return LineCollection(
				self.add("AXISX", Action.AC_BUTTON, action.x), self.add("AXISY", Action.AC_BUTTON, action.y),
			)
		line = Line(icon, action.describe(context))
		self.lines.append(line)
		return line

	def calculate(self, gen):
		self.width, self.height = self.min_width, 2 * self.PADDING
		self.icount = 0
		for line in self.lines:
			lw, lh = line.get_size(gen)
			self.width, self.height = max(self.width, lw), self.height + lh + self.SPACING
			self.icount = max(self.icount, len(line.icons))
		self.width += 2 * self.PADDING + self.icount * (gen.line_height + self.SPACING)
		self.width = min(self.width, self.max_width)
		self.height = max(self.height, self.min_height)

		anchor_x, anchor_y = self.anchor
		if (self.align & Align.TOP) != 0:
			self.y = anchor_y
		elif (self.align & Align.BOTTOM) != 0:
			self.y = gen.full_height - self.height - anchor_y
		else:
			self.y = (gen.full_height - self.height) / 2

		if (self.align & Align.LEFT) != 0:
			self.x = anchor_x
		elif (self.align & Align.RIGHT) != 0:
			self.x = gen.full_width - self.width - anchor_x
		else:
			self.x = (gen.full_width - self.width) / 2

	def place(self, gen, root):
		e = SVGEditor.add_element(
			root,
			"rect",
			style="opacity:1;fill-opacity:0.1;stroke-width:2.0;",
			fill="#00FF00",
			stroke="#06a400",
			id="box_%s" % (self.name,),
			width=self.width,
			height=self.height,
			x=self.x,
			y=self.y,
		)

		y = self.y + self.PADDING
		for line in self.lines:
			h = gen.line_height
			x = self.x + self.PADDING
			for icon in line.icons:
				image = find_image(icon)
				if image:
					# Fix: here stuff goes from weird to awfull, as rsvg
					# (library that gnome uses to render SVGs) can't render
					# linked images. Embeding is used instead.
					image = "data:image/svg+xml;base64,%s" % (base64.b64encode(open(image, "rb").read()))
					# Another problem: rsvg will NOT draw image unless href
					# tag uses namespace. No idea why is that, but I spent
					# 3 hours finding this, so I'm willing to murder.
					SVGEditor.add_element(
						root, "image", x=x, y=y, style="filter:url(#filterInvert)", width=h, height=h, href=image,
					)
				x += h + self.SPACING
			x = self.x + self.PADDING + self.icount * (h + self.SPACING)
			y += h
			txt = SVGEditor.add_element(root, "text", x=x, y=y, style=gen.label_template.attrib["style"])
			max_line_width = self.max_width - gen.line_height - self.PADDING
			while line.text and line.get_size(gen)[0] > max_line_width:
				line.text = line.text[:-1]
			SVGEditor.set_text(txt, line.text)
			y += self.SPACING

	def place_marker(self, gen, root):
		x1, y1 = self.x, self.y
		x2, y2 = x1 + self.width, y1 + self.height
		if self.align & (Align.LEFT | Align.RIGHT) == 0:
			edges = [[x2, y2], [x1, y2]]
		elif self.align & Align.BOTTOM == Align.BOTTOM:
			if self.align & Align.LEFT != 0:
				edges = [[x2, y2], [x1, y1]]
			elif self.align & Align.RIGHT != 0:
				edges = [[x2, y1], [x1, y2]]
		elif self.align & Align.TOP == Align.TOP:
			if self.align & Align.LEFT != 0:
				edges = [[x2, y1], [x2, y2]]
			elif self.align & Align.RIGHT != 0:
				edges = [[x1, y1], [x1, y2]]
		elif self.align & Align.LEFT != 0:
			edges = [[x2, y1], [x2, y2]]
		elif self.align & Align.RIGHT != 0:
			edges = [[x1, y1], [x2, y2]]

		targets = SVGEditor.get_element(root, "markers_%s" % (self.name,))
		if targets is None:
			return
		i = 0
		for target in targets:
			tx, ty = float(target.attrib["cx"]), float(target.attrib["cy"])
			try:
				edges[i] += [tx, ty]
				i += 1
			except IndexError:
				break
		edges = [i for i in edges if len(i) == 4]

		for x1, y1, x2, y2 in edges:
			e = SVGEditor.add_element(
				root,
				"line",
				style="opacity:1;stroke:#06a400;stroke-width:0.5;",
				# id = "box_%s_line0" % (self.name,),
				x1=x1,
				y1=y1,
				x2=x2,
				y2=y2,
			)


class Generator:
	PADDING = 10

	def __init__(self, editor, profile):
		background = SVGEditor.get_element(editor, "background")
		self.label_template = SVGEditor.get_element(editor, "label_template")
		self.line_height = int(float(self.label_template.attrib.get("height") or 8))
		self.char_width = int(float(self.label_template.attrib.get("width") or 8))
		self.full_width = int(float(background.attrib.get("width") or 800))
		self.full_height = int(float(background.attrib.get("height") or 800))

		boxes = []
		box_bcs = Box(0, self.PADDING, Align.TOP, "bcs")
		box_bcs.add("BACK", Action.AC_BUTTON, profile.buttons.get(SCButtons.BACK))
		box_bcs.add("C", Action.AC_BUTTON, profile.buttons.get(SCButtons.C))
		box_bcs.add("START", Action.AC_BUTTON, profile.buttons.get(SCButtons.START))
		boxes.append(box_bcs)

		box_left = Box(
			self.PADDING,
			self.PADDING,
			Align.LEFT | Align.TOP,
			"left",
			min_height=self.full_height * 0.5,
			min_width=self.full_width * 0.2,
			max_width=self.full_width * 0.275,
		)
		box_left.add("LEFT", Action.AC_TRIGGER, profile.triggers.get(profile.LEFT))
		box_left.add("LB", Action.AC_BUTTON, profile.buttons.get(SCButtons.LB))
		box_left.add("LGRIP", Action.AC_BUTTON, profile.buttons.get(SCButtons.LGRIP))
		box_left.add("LPAD", Action.AC_PAD, profile.pads.get(profile.LEFT))
		boxes.append(box_left)

		box_right = Box(
			self.PADDING,
			self.PADDING,
			Align.RIGHT | Align.TOP,
			"right",
			min_height=self.full_height * 0.5,
			min_width=self.full_width * 0.2,
			max_width=self.full_width * 0.275,
		)
		box_right.add("RIGHT", Action.AC_TRIGGER, profile.triggers.get(profile.RIGHT))
		box_right.add("RB", Action.AC_BUTTON, profile.buttons.get(SCButtons.RB))
		box_right.add("RGRIP", Action.AC_BUTTON, profile.buttons.get(SCButtons.RGRIP))
		box_right.add("RPAD", Action.AC_PAD, profile.pads.get(profile.RIGHT))
		boxes.append(box_right)

		box_abxy = Box(
			4 * self.PADDING, self.PADDING, Align.RIGHT | Align.BOTTOM, "abxy", max_width=self.full_width * 0.45,
		)
		box_abxy.add("A", Action.AC_BUTTON, profile.buttons.get(SCButtons.A))
		box_abxy.add("B", Action.AC_BUTTON, profile.buttons.get(SCButtons.B))
		box_abxy.add("X", Action.AC_BUTTON, profile.buttons.get(SCButtons.X))
		box_abxy.add("Y", Action.AC_BUTTON, profile.buttons.get(SCButtons.Y))
		boxes.append(box_abxy)

		box_stick = Box(
			4 * self.PADDING, self.PADDING, Align.LEFT | Align.BOTTOM, "stick", max_width=self.full_width * 0.45,
		)
		box_stick.add("STICK", Action.AC_STICK, profile.stick)
		boxes.append(box_stick)

		w = int(float(background.attrib.get("width") or 800))
		h = int(float(background.attrib.get("height") or 800))

		root = SVGEditor.get_element(editor, "root")
		for b in boxes:
			b.calculate(self)

		# Set ABXY and Stick size & position
		box_abxy.height = box_stick.height = self.full_height * 0.25
		box_abxy.width = box_stick.width = self.full_width * 0.3
		box_abxy.y = self.full_height - self.PADDING - box_abxy.height
		box_stick.y = self.full_height - self.PADDING - box_stick.height
		box_abxy.x = self.full_width - self.PADDING - box_abxy.width

		self.equal_width(box_left, box_right)
		self.equal_height(box_left, box_right)

		for b in boxes:
			b.place_marker(self, root)
		for b in boxes:
			b.place(self, root)

		editor.commit()

	def equal_width(self, *boxes):
		"""Sets width of all passed boxes to width of widest box"""
		width = 0
		for b in boxes:
			width = max(width, b.width)
		for b in boxes:
			b.width = width
			if b.align & Align.RIGHT:
				b.x = self.full_width - b.width - self.PADDING

	def equal_height(self, *boxes):
		"""Sets height of all passed boxes to height of tallest box"""
		height = 0
		for b in boxes:
			height = max(height, b.height)
		for b in boxes:
			b.height = height


def main():
	m = BindingDisplay()
	if not m.parse_argumets(sys.argv):
		sys.exit(1)
	m.run()
	sys.exit(m.get_exit_code())


if __name__ == "__main__":
	from scc.tools import init_logging

	init_logging()
	main()



"""SC-Controller - Paths.

Methods in this module are used to determine stuff like where user data is stored,
where sccdaemon can be executed from and similar.

This is gui-only thing, as sccdaemon doesn't really need to load anything what
python can't handle.
All this is needed since I want to have entire thing installable, runnable
from source tarball *and* debugable in working folder.
"""

import os
import sys


def get_config_path() -> str:
	"""Return configuration directory.

	~/.config/scc under normal conditions.
	"""
	confdir = os.path.expanduser("~/.config")
	if "XDG_CONFIG_HOME" in os.environ:
		confdir = os.environ["XDG_CONFIG_HOME"]
	return os.path.join(confdir, "scc")


def get_profiles_path() -> str:
	"""Return directory where profiles are stored.

	~/.config/scc/profiles under normal conditions.
	"""
	return os.path.join(get_config_path(), "profiles")


def get_default_profiles_path() -> str:
	"""Return directory where default profiles are stored.

	Probably something like /usr/share/scc/default_profiles,
	or $SCC_SHARED/default_profiles if program is being started from
	script extracted from source tarball
	"""
	return os.path.join(get_share_path(), "default_profiles")


def get_menuicons_path() -> str:
	"""Return directory where menu icons are stored.

	~/.config/scc/menu-icons under normal conditions.
	"""
	return os.path.join(get_config_path(), "menu-icons")


def get_default_menuicons_path() -> str:
	"""Return directory where default menu icons are stored.

	Probably something like /usr/share/scc/images/menu-icons,
	or $SCC_SHARED/images/menu-icons if program is being started from
	script extracted from source tarball
	"""
	return os.path.join(get_share_path(), "images/menu-icons")


def get_button_images_path() -> str:
	"""Return directory where button images are stored.

	/usr/share/scc/images/button-images by default.
	"""
	return os.path.join(get_share_path(), "images/button-images")


def get_menus_path() -> str:
	"""Return directory where profiles are stored.

	~/.config/scc/profiles under normal conditions.
	"""
	return os.path.join(get_config_path(), "menus")


def get_default_menus_path() -> str:
	"""Return directory where default profiles are stored.

	Probably something like /usr/share/scc/default_profiles,
	or ./default_profiles if program is being started from
	extracted source tarball
	"""
	return os.path.join(get_share_path(), "default_menus")


def get_controller_icons_path() -> str:
	"""Return directory where controller icons are stored.

	~/.config/scc/controller-icons under normal conditions.

	This directory may not exist.
	"""
	return os.path.join(get_config_path(), "controller-icons")


def get_default_controller_icons_path() -> str:
	"""Return directory where controller icons are stored.

	Probably something like /usr/share/scc/images/controller-icons,
	or ./images/controller-icons if program is being started from
	extracted source tarball.

	This directory should always exist.
	"""
	return os.path.join(get_share_path(), "images", "controller-icons")


def get_share_path() -> str:
	"""Return directory where shared files are kept.

	Usually "/usr/share/scc" or $SCC_SHARED if program is being started from
	script extracted from source tarball
	"""
	if "SCC_SHARED" in os.environ:
		return os.environ["SCC_SHARED"]
	paths = (
		"/usr/local/share/scc/",
		os.path.expanduser("~/.local/share/scc"),
		os.path.join(sys.prefix, "share/scc"),
	)
	for path in paths:
		if os.path.exists(path):
			return path
	# No path found, assume default and hope for best
	return "/usr/share/scc"


def get_pid_file() -> str:
	"""Return path to PID file.

	~/.config/scc/daemon.pid under normal conditions.
	"""
	return os.path.join(get_config_path(), "daemon.pid")


def get_daemon_socket() -> str:
	"""Return path to socket that can be used to control sccdaemon.

	~/.config/scc/daemon.socket under normal conditions.
	"""
	return os.path.join(get_config_path(), "daemon.socket")



#!/usr/bin/env python3
"""SC-Controller - Grid OSD Menu

Works as OSD menu, but displays item in (as rectangluar as possible - and
that's usually not very much) grid.
"""


from gi.repository import GObject, Gtk

from scc.config import Config
from scc.constants import CPAD, LEFT, RIGHT
from scc.gestures import GestureDetector
from scc.gui.daemon_manager import DaemonManager
from scc.gui.gestures import GestureDraw
from scc.osd import OSDWindow

BOTH = "BOTH"

import logging

log = logging.getLogger("osd.gesture")


class GestureDisplay(OSDWindow):
	"""OSD Window that displays gesture as it is being generated.

	Signals:
	  gesture-updated(gesture)		Emited repeadedly while gesture is being drawn.
	                                                                May be emited multiple times with same gesture.
	"""

	EPILOG = """Exit codes:
   0  - clean exit, user created gesture
  -1  - clean exit, user canceled gesture
   1  - error, invalid arguments
   2  - error, failed to access sc-daemon, sc-daemon reported error or died while menu is displayed.
   3  - erorr, failed to lock input
	"""
	__gsignals__ = {
		"gesture-updated": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
	}

	SIZE = 128  # times two horizontaly + borders

	def __init__(self, config=None):
		OSDWindow.__init__(self, "osd-gesture")
		self.daemon = None
		self._left_detector = GestureDetector(0, self._on_gesture_finished)
		# self._right_detector = GestureDetector(0, self._on_gesture_finished)
		self._control_with = LEFT
		self._eh_ids = []
		self._gesture = None

		self.setup_widgets()
		self.use_config(config or Config())

	def setup_widgets(self):
		self.parent = Gtk.Grid()
		self.parent.set_name("osd-gesture")

		self._left_draw = GestureDraw(self.SIZE, self._left_detector)
		# self._right_draw = GestureDraw(self.SIZE, self._right_detector)
		sep = Gtk.VSeparator()
		sep.set_name("osd-gesture-separator")

		self.parent.attach(self._left_draw, 0, 0, 1, 1)
		# self.parent.attach(sep,              1, 0, 1, 1)
		# self.parent.attach(self._right_draw, 2, 0, 1, 1)

		self.add(self.parent)

	def use_daemon(self, d):
		"""Allows (re)using already existing DaemonManager instance in same process.
		If this is used, parse_argumets() should be called before.
		"""
		self.daemon = d
		self.on_daemon_connected()

	def use_config(self, c):
		"""Allows reusing already existin Config instance in same process.
		Has to be called before parse_argumets()
		"""
		self.config = c
		# for x in (self._left_draw, self._right_draw):
		for x in (self._left_draw,):
			x.set_colors(**self.config["gesture_colors"])

	def _add_arguments(self):
		OSDWindow._add_arguments(self)
		self.argparser.add_argument(
			"--control-with",
			"-c",
			type=str,
			metavar="option",
			default=LEFT,
			choices=(LEFT, RIGHT, CPAD),
			help="which pad should be used to generate gesture menu (default: %s)" % (LEFT,),
		)

	def parse_argumets(self, argv):
		if not OSDWindow.parse_argumets(self, argv):
			return False
		if not self.config:
			self.use_config(Config())

		# Parse simpler arguments
		self._control_with = self.args.control_with

		return True

	def _connect_handlers(self):
		self._eh_ids += [
			(self.daemon, self.daemon.connect("dead", self.on_daemon_died)),
			(self.daemon, self.daemon.connect("error", self.on_daemon_died)),
			(self.daemon, self.daemon.connect("alive", self.on_daemon_connected)),
		]

	def run(self):
		self.daemon = DaemonManager()
		self._connect_handlers()
		OSDWindow.run(self)

	def show(self, *a):
		OSDWindow.show(self, *a)

	def on_daemon_connected(self, *a):
		def success(*a):
			log.error("Sucessfully locked %s pad", self._control_with)
			self._left_detector.enable()
			# self._right_detector.enable()

		if not self.config:
			self.config = Config()
		locks = [self._control_with]
		c = self.choose_controller(self.daemon)
		if c is None or not c.is_connected():
			# There is no controller connected to daemon
			self.on_failed_to_lock("Controller not connected")
			return

		self._eh_ids += [
			(c, c.connect("event", self.on_event)),
			(c, c.connect("lost", self.on_controller_lost)),
		]
		c.lock(success, self.on_failed_to_lock, *locks)

	def quit(self, code=-2):
		if self.get_controller():
			self.get_controller().unlock_all()
		for source, eid in self._eh_ids:
			source.disconnect(eid)
		self._eh_ids = []
		OSDWindow.quit(self, code)

	def on_event(self, daemon, what, data):
		if what == self._control_with:
			x, y = data
			self._left_draw.add(x, y)
			self._left_detector.whole(None, x, y, what)
			# TODO: self._right_detector, if there is any use for it later
			self.emit("gesture-updated", self._left_detector.get_string())

	def get_gesture(self):
		"""Returns recognized gesture or None if there is not any"""
		if self._gesture:
			return self._gesture
		# self._gesture is None or empty
		return None

	def _on_gesture_finished(self, detector, gesture):
		self._gesture = gesture
		log.debug("Recognized gesture: %s", gesture)
		self.quit(0)


def main():
	import gi

	gi.require_version("Gtk", "3.0")
	gi.require_version("Rsvg", "2.0")
	gi.require_version("GdkX11", "3.0")

	from scc.tools import init_logging

	init_logging()

	gd = GestureDisplay()
	if not gd.parse_argumets(sys.argv):
		sys.exit(1)
	gd.run()
	if gd.get_exit_code() == 0:
		print(gd.get_gesture())
	else:
		sys.exit(gd.get_exit_code())


if __name__ == "__main__":
	import signal
	import sys

	def sigint(*a):
		print("\n*break*")
		sys.exit(-1)

	signal.signal(signal.SIGINT, sigint)
	main()



#!/usr/bin/env python3
"""SC-Controller - GDK_TO_KEY

Maps Gdk.KEY_* constants into Keys.KEY_* constants.
Used by ActionEditor (when grabbing the key)
"""

from gi.repository import Gdk

from scc.uinput import Keys

GDK_TO_KEY = {
	# Row 1
	Gdk.KEY_Escape: Keys.KEY_ESC,
	Gdk.KEY_Print: Keys.KEY_PRINT,
	Gdk.KEY_Scroll_Lock: Keys.KEY_SCROLLLOCK,
	Gdk.KEY_Sys_Req: Keys.KEY_SYSRQ,
	Gdk.KEY_Pause: Keys.KEY_PAUSE,
	# Row 2
	Gdk.KEY_quoteleft: Keys.KEY_GRAVE,  # tilde
	Gdk.KEY_minus: Keys.KEY_MINUS,
	Gdk.KEY_equal: Keys.KEY_EQUAL,
	Gdk.KEY_BackSpace: Keys.KEY_BACKSPACE,
	# Row 3
	Gdk.KEY_Tab: Keys.KEY_TAB,
	Gdk.KEY_bracketleft: Keys.KEY_LEFTBRACE,
	Gdk.KEY_bracketright: Keys.KEY_RIGHTBRACE,
	Gdk.KEY_backslash: Keys.KEY_BACKSLASH,
	# Row 4
	Gdk.KEY_Caps_Lock: Keys.KEY_CAPSLOCK,
	Gdk.KEY_semicolon: Keys.KEY_SEMICOLON,
	Gdk.KEY_apostrophe: Keys.KEY_APOSTROPHE,
	Gdk.KEY_Return: Keys.KEY_ENTER,
	# Row 5
	Gdk.KEY_Shift_L: Keys.KEY_LEFTSHIFT,
	Gdk.KEY_comma: Keys.KEY_COMMA,
	Gdk.KEY_period: Keys.KEY_DOT,
	Gdk.KEY_slash: Keys.KEY_SLASH,
	Gdk.KEY_Shift_R: Keys.KEY_RIGHTSHIFT,
	# Numpad
	Gdk.KEY_KP_0: Keys.KEY_KP0,
	Gdk.KEY_KP_1: Keys.KEY_KP1,
	Gdk.KEY_KP_2: Keys.KEY_KP2,
	Gdk.KEY_KP_3: Keys.KEY_KP3,
	Gdk.KEY_KP_4: Keys.KEY_KP4,
	Gdk.KEY_KP_5: Keys.KEY_KP5,
	Gdk.KEY_KP_6: Keys.KEY_KP6,
	Gdk.KEY_KP_7: Keys.KEY_KP7,
	Gdk.KEY_KP_8: Keys.KEY_KP8,
	Gdk.KEY_KP_9: Keys.KEY_KP9,
	Gdk.KEY_KP_Delete: Keys.KEY_KPDOT,
	Gdk.KEY_KP_Divide: Keys.KEY_KPSLASH,
	Gdk.KEY_KP_Add: Keys.KEY_KPPLUS,
	Gdk.KEY_KP_Multiply: Keys.KEY_KPASTERISK,
	Gdk.KEY_KP_Subtract: Keys.KEY_KPMINUS,
	Gdk.KEY_KP_Enter: Keys.KEY_KPENTER,
	Gdk.KEY_Num_Lock: Keys.KEY_NUMLOCK,
	# Home & co
	Gdk.KEY_Insert: Keys.KEY_INSERT,
	Gdk.KEY_Home: Keys.KEY_HOME,
	Gdk.KEY_Page_Up: Keys.KEY_PAGEUP,
	Gdk.KEY_Delete: Keys.KEY_DELETE,
	Gdk.KEY_End: Keys.KEY_END,
	Gdk.KEY_Page_Down: Keys.KEY_PAGEDOWN,
	# Arrows
	Gdk.KEY_Up: Keys.KEY_UP,
	Gdk.KEY_Left: Keys.KEY_LEFT,
	Gdk.KEY_Right: Keys.KEY_RIGHT,
	Gdk.KEY_Down: Keys.KEY_DOWN,
	# Bottom row
	Gdk.KEY_Control_L: Keys.KEY_LEFTCTRL,
	Gdk.KEY_Super_L: Keys.KEY_LEFTMETA,
	Gdk.KEY_Alt_L: Keys.KEY_LEFTALT,
	Gdk.KEY_space: Keys.KEY_SPACE,
	Gdk.KEY_Alt_R: Keys.KEY_RIGHTALT,
	Gdk.KEY_Super_R: Keys.KEY_RIGHTMETA,
	Gdk.KEY_Menu: Keys.KEY_COMPOSE,
	Gdk.KEY_Control_R: Keys.KEY_RIGHTCTRL,
}


KEYCODE_TO_KEY = {
	# Row 1
	9: Keys.KEY_ESC,
	67: Keys.KEY_F1,
	68: Keys.KEY_F2,
	69: Keys.KEY_F3,
	70: Keys.KEY_F4,
	71: Keys.KEY_F5,
	72: Keys.KEY_F6,
	73: Keys.KEY_F7,
	74: Keys.KEY_F8,
	75: Keys.KEY_F9,
	76: Keys.KEY_F10,
	95: Keys.KEY_F11,
	96: Keys.KEY_F12,
	107: Keys.KEY_PRINT,
	78: Keys.KEY_SCROLLLOCK,
	127: Keys.KEY_PAUSE,
	# Row 2
	49: Keys.KEY_GRAVE,  # tilde
	10: Keys.KEY_1,
	11: Keys.KEY_2,
	12: Keys.KEY_3,
	13: Keys.KEY_4,
	14: Keys.KEY_5,
	15: Keys.KEY_6,
	16: Keys.KEY_7,
	17: Keys.KEY_8,
	18: Keys.KEY_9,
	19: Keys.KEY_0,
	20: Keys.KEY_MINUS,
	21: Keys.KEY_EQUAL,
	22: Keys.KEY_BACKSPACE,
	# Row 3
	23: Keys.KEY_TAB,
	24: Keys.KEY_Q,
	25: Keys.KEY_W,
	26: Keys.KEY_E,
	27: Keys.KEY_R,
	28: Keys.KEY_T,
	29: Keys.KEY_Y,
	30: Keys.KEY_U,
	31: Keys.KEY_I,
	32: Keys.KEY_O,
	33: Keys.KEY_P,
	34: Keys.KEY_LEFTBRACE,
	35: Keys.KEY_RIGHTBRACE,
	51: Keys.KEY_BACKSLASH,
	# Row 4
	66: Keys.KEY_CAPSLOCK,
	38: Keys.KEY_A,
	39: Keys.KEY_S,
	40: Keys.KEY_D,
	41: Keys.KEY_F,
	42: Keys.KEY_G,
	43: Keys.KEY_H,
	44: Keys.KEY_J,
	45: Keys.KEY_K,
	46: Keys.KEY_L,
	47: Keys.KEY_SEMICOLON,
	48: Keys.KEY_APOSTROPHE,
	36: Keys.KEY_ENTER,
	# Row 5
	50: Keys.KEY_LEFTSHIFT,
	52: Keys.KEY_Z,
	53: Keys.KEY_X,
	54: Keys.KEY_C,
	55: Keys.KEY_V,
	56: Keys.KEY_B,
	57: Keys.KEY_N,
	58: Keys.KEY_M,
	59: Keys.KEY_COMMA,
	60: Keys.KEY_DOT,
	61: Keys.KEY_SLASH,
	62: Keys.KEY_RIGHTSHIFT,
	# Numpad
	90: Keys.KEY_KP0,
	87: Keys.KEY_KP1,
	88: Keys.KEY_KP2,
	89: Keys.KEY_KP3,
	83: Keys.KEY_KP4,
	84: Keys.KEY_KP5,
	85: Keys.KEY_KP6,
	79: Keys.KEY_KP7,
	80: Keys.KEY_KP8,
	81: Keys.KEY_KP9,
	91: Keys.KEY_KPDOT,
	106: Keys.KEY_KPSLASH,
	86: Keys.KEY_KPPLUS,
	63: Keys.KEY_KPASTERISK,
	82: Keys.KEY_KPMINUS,
	104: Keys.KEY_KPENTER,
	77: Keys.KEY_NUMLOCK,
	# Home & co
	118: Keys.KEY_INSERT,
	110: Keys.KEY_HOME,
	112: Keys.KEY_PAGEUP,
	119: Keys.KEY_DELETE,
	115: Keys.KEY_END,
	117: Keys.KEY_PAGEDOWN,
	# Arrows
	111: Keys.KEY_UP,
	113: Keys.KEY_LEFT,
	114: Keys.KEY_RIGHT,
	116: Keys.KEY_DOWN,
	# Bottom row
	37: Keys.KEY_LEFTCTRL,
	133: Keys.KEY_LEFTMETA,
	64: Keys.KEY_LEFTALT,
	65: Keys.KEY_SPACE,
	108: Keys.KEY_RIGHTALT,
	134: Keys.KEY_RIGHTMETA,
	135: Keys.KEY_COMPOSE,
	105: Keys.KEY_RIGHTCTRL,
}

# Stuff that is missing above is auto-generated here
names = {x.name: x for x in Keys}

for x in dir(Gdk):
	if x.startswith("KEY_"):
		if x in names:
			GDK_TO_KEY[getattr(Gdk, x)] = names[x]

# A-Z keys, because GDK has different codes for 'A' and 'a'
for x in range(ord("a"), ord("z") + 1):
	GDK_TO_KEY[getattr(Gdk, "KEY_" + chr(x))] = names["KEY_" + chr(x).upper()]

KEY_TO_GDK = {GDK_TO_KEY[a]: a for a in GDK_TO_KEY}
KEY_TO_KEYCODE = {KEYCODE_TO_KEY[a]: a for a in KEYCODE_TO_KEY}


def keyevent_to_key(event):
	if event.hardware_keycode in KEYCODE_TO_KEY:
		return KEYCODE_TO_KEY[event.hardware_keycode]
	return None



"""Python wrapper for some X-related stuff.

Copyright (C) 2017 Kozec

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2 as published by
the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from ctypes import (
	CDLL,
	POINTER,
	Structure,
	byref,
	c_bool,
	c_char_p,
	c_int,
	c_long,
	c_short,
	c_ubyte,
	c_uint,
	c_ulong,
	c_ushort,
	c_void_p,
	cast,
)


def _load_lib(*names):
	"""Tries multiple alternative names to load .so library.
	"""
	for l in names:
		try:
			return CDLL(l)
		except OSError:
			pass
	raise OSError("Failed to load %s, library not found" % (names[0],))


libXFixes = _load_lib("libXfixes.so", "libXfixes.so.3")
libX11 = _load_lib("libX11.so", "libX11.so.6")
libXext = _load_lib("libXext.so", "libXext.so.6")


# Types
XID = c_ulong
Pixmap = XID
Colormap = XID
Atom = c_ulong
XserverRegion = c_ulong
GC = c_void_p
Display = c_void_p


# Structures
class XRectangle(Structure):
	_fields_ = [
		("x", c_short),
		("y", c_short),
		("width", c_ushort),
		("height", c_ushort),
	]


class XClassHint(Structure):
	_fields_ = [
		("res_name", c_char_p),
		("res_class", c_char_p),
	]


class XkbStateRec(Structure):
	_fields_ = [
		("group", c_ubyte),
		("locked_group", c_ubyte),
		("base_group", c_ushort),
		("latched_group", c_ushort),
		("mods", c_ubyte),
		("base_mods", c_ubyte),
		("latched_mods", c_ubyte),
		("locked_mods", c_ubyte),
		("compat_state", c_ubyte),
		("grab_mods", c_ubyte),
		("compat_grab_mods", c_ubyte),
		("lookup_mods", c_ubyte),
		("compat_lookup_mods", c_ubyte),
		("ptr_buttons", c_ushort),
	]


class XWindowAttributes(Structure):
	_fields_ = [
		("x", c_int),
		("y", c_int),
		("width", c_int),
		("height", c_int),
		("depth", c_int),
		("visual", c_void_p),
		("root", XID),
		("i_class", c_int),
		("bit_gravity", c_int),
		("win_gravity", c_int),
		("backing_store", c_int),
		("backing_planes", c_ulong),
		("backing_pixel", c_ulong),
		("save_under", c_bool),
		("colormap", Colormap),
		("map_installed", c_bool),
		("map_state", c_int),
		("all_event_masks", c_long),
		("your_event_mask", c_long),
		("do_not_propagate_mask", c_long),
		("map_installed", c_bool),
		("screen", c_void_p),
	]


# Consants
SHAPE_BOUNDING = 0
SHAPE_CLIP = 1
SHAPE_INPUT = 2
SHAPE_SET = 0

XKBUSECOREKBD = 0x0100
ANYPROPERTYTYPE = 0
SUCCESS = 0

ISVIEWABLE = 2


# Functions
open_display = libX11.XOpenDisplay
open_display.__doc__ = "Opens connection to XDisplay"
open_display.argtypes = [c_char_p]
open_display.restype = c_void_p

free = libX11.XFree
free.__doc__ = "Used to free some resource returned by XLib"
free.argtypes = [c_void_p]

create_region = libXFixes.XFixesCreateRegion
create_region.__doc__ = "Creates rectanglular region for use with set_window_shape_region"
create_region.argtypes = [c_void_p, POINTER(XRectangle), c_int]
create_region.restype = XserverRegion

set_window_shape_region = libXFixes.XFixesSetWindowShapeRegion
set_window_shape_region.__doc__ = "Sets region in which window accepts inputs"
set_window_shape_region.argtypes = [c_void_p, XID, c_int, c_int, c_int, XserverRegion]

destroy_region = libXFixes.XFixesDestroyRegion
destroy_region.__doc__ = "Frees region created by create_region"
destroy_region.argtypes = [c_void_p, XserverRegion]

get_default_root_window = libX11.XDefaultRootWindow
get_default_root_window.argtypes = [c_void_p]

flush = libX11.XFlush
flush.__doc__ = "Asks Xlib to send queued commands to XServer"
flush.argtypes = [c_void_p]

warp_pointer = libX11.XWarpPointer
warp_pointer.__doc__ = "Very, very, V*E*R*Y complicated shit used to move cursor"
warp_pointer.argtypes = [c_void_p, XID, XID, c_int, c_int, c_int, c_int, c_int, c_int]

query_pointer = libX11.XQueryPointer
query_pointer.__doc__ = "Returns a lot of nonsense along with mouse cursor position"
query_pointer.argtypes = [
	c_void_p,
	XID,
	POINTER(XID),
	POINTER(XID),
	POINTER(c_int),
	POINTER(c_int),
	POINTER(c_int),
	POINTER(c_int),
	POINTER(c_uint),
]

get_window_attributes = libX11.XGetWindowAttributes
get_window_attributes.__doc__ = "https://tronche.com/gui/x/xlib/window-information/XGetWindowAttributes.html"
get_window_attributes.argtypes = [c_void_p, XID, POINTER(XWindowAttributes)]


translate_coordinates = libX11.XTranslateCoordinates
translate_coordinates.argtypes = [c_void_p, XID, XID, c_int, c_int, POINTER(c_int), POINTER(c_int), POINTER(XID)]
translate_coordinates.restype = c_bool


get_input_focus = libX11.XGetInputFocus
get_input_focus.__doc__ = """Returns window that currently have window focus.
	Most of window managers and some GTK applications are breaking this.
	See https://specifications.freedesktop.org/wm-spec/1.3/ar01s03.html
	"""
get_input_focus.argtypes = [c_void_p, POINTER(XID), POINTER(c_int)]

get_window_property = libX11.XGetWindowProperty
get_window_property.__doc__ = "Returns value of property associated with window"
get_window_property.argtypes = [
	c_void_p,
	XID,
	Atom,
	c_long,
	c_long,
	c_bool,
	Atom,
	POINTER(Atom),
	POINTER(Atom),
	POINTER(c_ulong),
	POINTER(c_ulong),
	POINTER(c_void_p),
]
get_window_property.restype = c_int

alloc_class_hint = libX11.XAllocClassHint
alloc_class_hint.restype = POINTER(XClassHint)
alloc_class_hint.argtypes = []
alloc_class_hint.__doc__ = """Allocates and returns a pointer to a XClassHint
	structure. Returned pointer has to be deallocated using free()"""

get_class_hint = libX11.XGetClassHint
get_class_hint.argtypes = [c_void_p, XID, POINTER(XClassHint)]
get_class_hint.restype = c_int

intern_atom = libX11.XInternAtom
intern_atom.__doc__ = "Returns integer ID for specified Atom name."
intern_atom.argtypes = [c_void_p, c_char_p, c_bool]
intern_atom.restype = Atom

create_pixmap = libX11.XCreatePixmap
create_pixmap.argtypes = [c_void_p, XID, c_uint, c_uint, c_uint]
create_pixmap.restype = Pixmap

create_pixmap_from_bitmap = libX11.XCreatePixmapFromBitmapData
create_pixmap_from_bitmap.argtypes = [c_void_p, XID, c_char_p, c_uint, c_uint, c_uint, c_uint, c_uint]
create_pixmap_from_bitmap.restype = Pixmap

write_bitmap = libX11.XWriteBitmapFile
write_bitmap.argtypes = [c_void_p, c_char_p, Pixmap, c_uint, c_uint, c_int, c_int]
write_bitmap.restype = c_int

free_pixmap = libX11.XFreePixmap
free_pixmap.__doc__ = "Deallocates pixmap created by create_pixmap"
free_pixmap.argtypes = [c_void_p, Pixmap]

create_gc = libX11.XCreateGC
create_gc.__doc__ = "Creates graphics context to draw on"
create_gc.argtypes = [c_void_p, XID, c_ulong, c_void_p]
create_gc.restype = GC

flush_gc = libX11.XFlushGC
flush_gc.__doc__ = "Force sending GC component changes"
flush_gc.argtypes = [c_void_p, GC]

free_gc = libX11.XFreeGC
free_gc.__doc__ = "Deallocates graphics context created by create_gc"
free_gc.argtypes = [c_void_p, GC]

fill_rectangle = libX11.XFillRectangle
fill_rectangle.__doc__ = "Draws and fills rectangle on graphics context"
fill_rectangle.argtypes = [c_void_p, XID, GC, c_int, c_int, c_uint, c_uint]

draw_arc = libX11.XDrawArc
draw_arc.argtypes = [c_void_p, Pixmap, GC, c_int, c_int, c_uint, c_uint, c_int, c_int]

fill_arc = libX11.XFillArc
fill_arc.argtypes = [c_void_p, Pixmap, GC, c_int, c_int, c_uint, c_uint, c_int, c_int]


set_foreground = libX11.XSetForeground
set_foreground.__doc__ = "Sets foreground color for drawing on graphics context"
set_foreground.argtypes = [c_void_p, GC, c_ulong]

set_background = libX11.XSetBackground
set_background.__doc__ = "Sets background color for drawing on graphics context"
set_background.argtypes = set_foreground.argtypes

shape_combine_mask = libXext.XShapeCombineMask
shape_combine_mask.__doc__ = "Sets 1-bit transparency mask for window"
shape_combine_mask.argtypes = [c_void_p, XID, c_int, c_int, c_int, Pixmap, c_int]


# Wrapped functions
_xkb_get_state = libX11.XkbGetState
_xkb_get_state.argtypes = [c_void_p, c_uint, POINTER(XkbStateRec)]


# Wrappers
def get_xkb_state(dpy):
	rec = XkbStateRec()
	_xkb_get_state(dpy, XKBUSECOREKBD, rec)
	return rec


def get_window_size(dpy, window):
	attrs = XWindowAttributes()
	get_window_attributes(dpy, window, byref(attrs))
	return attrs.width, attrs.height


def is_window_visible(dpy, window):
	"""Return True if window mapping state is IsViewable"""
	attrs = XWindowAttributes()
	get_window_attributes(dpy, window, byref(attrs))
	return attrs.map_state == ISVIEWABLE


def get_window_geometry(dpy, win):
	"""Returns window x,y,width,height"""
	attrs = XWindowAttributes()
	get_window_attributes(dpy, win, byref(attrs))
	x, y = c_int(), c_int()
	trash = XID()
	if translate_coordinates(dpy, win, get_default_root_window(dpy), 0, 0, byref(x), byref(y), byref(trash)):
		return x.value, y.value, attrs.width, attrs.height
	# translate_coordinates failed
	return attrs.x, attrs.y, attrs.width, attrs.height


def get_screen_size(dpy):
	return get_window_size(dpy, get_default_root_window(dpy))


def get_mouse_pos(dpy, relative_to=None):
	"""Returns mouse position relative to specified window or to screen, if no
	window is specified.
	"""
	if relative_to is None:
		relative_to = get_default_root_window(dpy)
	root_return, child = XID(), XID()
	x, y = c_int(), c_int()
	child_x, child_y = c_int(), c_int()
	mask = c_uint()

	query_pointer(
		dpy,
		relative_to,
		byref(root_return),
		byref(child),
		byref(x),
		byref(y),
		byref(child_x),
		byref(child_y),
		byref(mask),
	)
	return x.value, y.value


def set_mouse_pos(dpy, x, y, relative_to=None):
	"""Sets mouse position relative to specified window or to screen, if no
	window is specified.
	"""
	if relative_to is None:
		relative_to = get_default_root_window(dpy)
	warp_pointer(dpy, 0, relative_to, 0, 0, 0, 0, x, y)
	flush(dpy)


def get_window_prop(dpy, window, prop_name, max_size=2):
	"""Returns (nitems, property) of specified window or (-1, None) if anything fails.
	Returned 'property' is POINTER(c_void_p) and has to be freed using X.free().
	"""
	if type(prop_name) is str:
		prop_name = prop_name.encode("utf-8")

	prop_atom = intern_atom(dpy, prop_name, False)
	type_return, format_return = Atom(), Atom()
	nitems, bytes_after = c_ulong(), c_ulong()
	prop = c_void_p()

	if get_window_property(
		dpy,
		window,
		prop_atom,
		0,
		max_size,
		False,
		ANYPROPERTYTYPE,
		byref(type_return),
		byref(format_return),
		byref(nitems),
		byref(bytes_after),
		byref(prop),
	) == SUCCESS:
		return nitems.value, prop
	return -1, None


def get_current_window(dpy):
	"""Returns active window or root window if there is no active.
	"""
	# Try using WM-provided info first
	trash, prop = get_window_prop(dpy, get_default_root_window(dpy), "_NET_ACTIVE_WINDOW")
	if prop is not None:
		rv = cast(prop, POINTER(Atom)).contents.value
		free(prop)
		return rv

	# Fall-back to something what probably can't work anyway
	win, revert_to = XID(), c_int()
	get_input_focus(dpy, byref(win), byref(revert_to))
	if win == 0:
		return get_default_root_window(dpy)
	return win


def get_window_type(dpy, window):
	"""Returns _NET_WM_WINDOW_TYPE value for window specified or None if anything
	fails while receiving it.
	"""
	trash, prop = get_window_prop(dpy, window, "_NET_WM_WINDOW_TYPE")
	if prop is not None and prop.value is not None:
		rv = cast(prop, POINTER(Atom)).contents.value
		free(prop)
		return rv
	return None


def get_window_title(dpy, window):
	"""Returns window title or None if title cannot be obtained.
	"""
	for prop_name in ("_NET_WM_NAME", "WM_NAME"):
		trash, prop = get_window_prop(dpy, window, prop_name, max_size=2048)
		if prop:
			try:
				value = cast(prop, c_char_p).value.decode("utf-8")
				free(prop)
				return value
			except:
				pass
			free(prop)
	return None


def get_window_class(dpy, window):
	"""Returns window class or None, None if class cannot be obtained.
	"""
	s = alloc_class_hint()
	if s:
		if get_class_hint(dpy, window, s):
			value = s.contents.res_name.decode("utf-8"), s.contents.res_class.decode("utf-8")
			free(s)
			return value
		free(s)

	return None, None


def get_wm_state(dpy, window):
	"""Returns list of _NET_WM_STATE atoms assotiated with window or empty list
	if list be obtained.
	"""
	count, state = get_window_prop(dpy, window, "_NET_WM_STATE", 1024)
	if count <= 0:
		return []
	return cast(state, POINTER(Atom))[0:count]



#!/usr/bin/env python3
"""SC-Controller - Action Editor - common part of "DPAD or menu" and "Special Action",
two components with MenuAction selectable.
"""

import logging
import os
from typing import Self

from gi.repository import Gtk

from scc.actions import NoAction
from scc.constants import DEFAULT, SAME, STICK, SCButtons
from scc.gui.menu_editor import MenuEditor
from scc.gui.parser import GuiActionParser
from scc.gui.userdata_manager import UserDataManager
from scc.paths import get_menus_path
from scc.special_actions import (
	GridMenuAction,
	HorizontalMenuAction,
	MenuAction,
	PositionModifier,
	QuickMenuAction,
	RadialMenuAction,
)
from scc.tools import _, nameof

log = logging.getLogger("AE.Menu")

__all__ = ["MenuActionCofC"]


class MenuActionCofC(UserDataManager):
	# CofC - Component of Component
	def __init__(self):
		UserDataManager.__init__(self)
		self._current_menu = None
		self.parser = GuiActionParser()
		self.allow_globals = True
		self.allow_in_profile = True

	def allow_menus(self, allow_globals, allow_in_profile) -> Self:
		"""Sets which type of menu should be selectable.
		By default, both are enabled.

		Returns self.
		"""
		self.allow_globals = allow_globals
		self.allow_in_profile = allow_in_profile
		return self

	def set_selected_menu(self, menu):
		"""Sets menu selected in combobox.
		Returns self.
		"""
		self._current_menu = menu
		# TODO: This currently works only if menu list is not yet loaded

	@staticmethod
	def menu_class_to_key(action):
		"""For subclass of MenuAction, returns correct key to be used in ListStore.
		"""
		if isinstance(action, GridMenuAction):
			return "gridmenu"
		if isinstance(action, QuickMenuAction):
			return "quickmenu"
		if isinstance(action, HorizontalMenuAction):
			return "hmenu"
		if isinstance(action, RadialMenuAction):
			return "radialmenu"
		return "menu"

	def load_menu_data(self, action):
		if isinstance(action, PositionModifier):
			# Load menu position modifier, if used
			x, y = action.position
			self.builder.get_object("cbMenuPosX").set_active(0 if x >= 0 else 1)
			self.builder.get_object("cbMenuPosY").set_active(0 if y >= 0 else 1)
			self.builder.get_object("spMenuPosX").set_value(abs(x))
			self.builder.get_object("spMenuPosY").set_value(abs(y))
			action = action.action

		self._current_menu = action.menu_id
		cbm = self.builder.get_object("cbMenuType")
		self.set_cb(cbm, self.menu_class_to_key(action), 1)

		if self.builder.get_object("rvMenuSize"):
			spMenuSize = self.builder.get_object("spMenuSize")
			if self.update_size_display(action):
				size = spMenuSize.get_adjustment().set_value(action.size)

		cbControlWith = self.builder.get_object("cbControlWith")
		cbConfirmWith = self.builder.get_object("cbConfirmWith")
		cbCancelWith = self.builder.get_object("cbCancelWith")
		cbMenuAutoConfirm = self.builder.get_object("cbMenuAutoConfirm")
		cbMenuConfirmWithClick = self.builder.get_object("cbMenuConfirmWithClick")
		cbMenuAutoCancel = self.builder.get_object("cbMenuAutoCancel")
		if cbControlWith:
			self.set_cb(cbControlWith, nameof(action.control_with), 1)

		cow = action.confirm_with
		caw = action.cancel_with

		if cbConfirmWith:
			if cow == SAME and cbMenuAutoConfirm:
				cbMenuAutoConfirm.set_active(True)
				cbConfirmWith.set_sensitive(False)
			elif cbMenuConfirmWithClick and cow == self.get_default_confirm():
				cbMenuConfirmWithClick.set_active(True)
				cbConfirmWith.set_sensitive(False)
			else:
				if cbMenuAutoConfirm:
					cbMenuAutoConfirm.set_active(False)
				if cbMenuConfirmWithClick:
					cbMenuAutoConfirm.set_active(False)
				cbConfirmWith.set_sensitive(True)
				self.set_cb(cbConfirmWith, nameof(cow), 1)

		if cbCancelWith:
			if caw == SAME and cbMenuAutoCancel:
				cbMenuAutoCancel.set_active(True)
				cbCancelWith.set_sensitive(False)
			else:
				if cbMenuAutoCancel:
					cbMenuAutoCancel.set_active(False)
				self.set_cb(cbCancelWith, nameof(caw), 1)

		self.on_cbMenus_changed()

	def get_default_confirm(self):
		"""Returns DEFAULT, but may be overriden when default
		confirm button is different - specifically when used with pads.
		"""
		return DEFAULT

	def get_default_cancel(self):
		"""Returns DEFAULT, but may be overriden when default
		confirm button is different - specifically when used with pads.
		"""
		return DEFAULT

	def on_menu_changed(self, new_id):
		self._current_menu = new_id
		self.editor.set_action(MenuAction(new_id))
		self.load_menu_list()

	def on_btEditMenu_clicked(self, *a):
		name = self.get_selected_menu()
		if name:
			log.debug("Editing %s", name)
			me = MenuEditor(self.app, self.on_menu_changed)
			id = self.get_selected_menu()
			log.debug("Opening editor for menu ID '%s'", id)
			me.set_menu(id)
			me.allow_menus(self.allow_globals, self.allow_in_profile)
			me.show(self.editor.window)

	def on_cbMenus_button_press_event(self, trash, event):
		if event.button == 3:
			mnuMenu = self.builder.get_object("mnuMenu")
			mnuMenu.popup(None, None, None, None, 3, Gtk.get_current_event_time())

	def on_mnuMenuNew_activate(self, *a):
		self.on_new_menu_selected()

	def on_mnuMenuCopy_activate(self, *a):
		self.on_new_menu_selected(make_copy=True)

	def on_mnuMenuRename_activate(self, *a):
		self.on_btEditMenu_clicked()

	def on_mnuMenuDelete_activate(self, *a):
		id = self.get_selected_menu()
		if MenuEditor.menu_is_global(id):
			text = _("Really delete selected global menu?")
		else:
			text = _("Really delete selected menu?")

		d = Gtk.MessageDialog(
			parent=self.editor.window,
			flags=Gtk.DialogFlags.MODAL,
			type=Gtk.MessageType.WARNING,
			buttons=Gtk.ButtonsType.OK_CANCEL,
			message_format=text,
		)

		if MenuEditor.menu_is_global(id):
			d.format_secondary_text(_("This action is not undoable!"))

		if d.run() == -5:  # OK button, no idea where is this defined...
			if MenuEditor.menu_is_global(id):
				fname = os.path.join(get_menus_path(), id)
				try:
					os.unlink(fname)
				except Exception as e:
					log.error("Failed to remove %s: %s", fname, e)
			else:
				del self.app.current.menus[id]
				self.app.on_profile_modified()
			self.load_menu_list()
		d.destroy()

	def on_menus_loaded(self, menus):
		cb = self.builder.get_object("cbMenus")
		cb.set_row_separator_func(lambda model, iter: model.get_value(iter, 1) is None)
		model = cb.get_model()
		model.clear()
		i, current_index = 0, 0
		if self.allow_in_profile:
			# Add menus from profile
			for key in sorted(self.app.current.menus):
				model.append((key, key))
				if self._current_menu == key:
					current_index = i
				i += 1
			if i > 0:
				model.append((None, None))  # Separator
				i += 1
		if self.allow_globals:
			for f in menus:
				key = f.get_basename()
				name = key
				if name.startswith("."):
					continue
				if "." in name:
					name = _("%s (global)" % (name.split(".")[0]))
				model.append((name, key))
				if self._current_menu == key:
					current_index = i
				i += 1
		if i > 0:
			model.append((None, None))  # Separator
		model.append((_("New Menu..."), ""))

		self._recursing = True
		cb.set_active(current_index)
		self._recursing = False
		name = self.get_selected_menu()
		if name:
			self.builder.get_object("btEditMenu").set_sensitive(name not in MenuEditor.OPEN)

	def handles(self, mode, action):
		if isinstance(action, PositionModifier):
			return isinstance(action.action, MenuAction)
		return isinstance(action, MenuAction)

	def get_selected_menu(self):
		cb = self.builder.get_object("cbMenus")
		model = cb.get_model()
		iter = cb.get_active_iter()
		if iter is None:
			# Empty list
			return None
		return model.get_value(iter, 1)

	def prevent_confirm_cancel_nonsense(self, widget, *a):
		"""If 'confirm with click', 'confirm with release' and
		'cbMenuAutoCancel' are all present, this method prevents them from
		being checked in nonsensical way.
		"""
		cbMenuConfirmWithClick = self.builder.get_object("cbMenuConfirmWithClick")
		cbMenuAutoConfirm = self.builder.get_object("cbMenuAutoConfirm")
		cbMenuAutoCancel = self.builder.get_object("cbMenuAutoCancel")
		if widget.get_active():
			if widget == cbMenuConfirmWithClick:
				if cbMenuAutoConfirm:
					cbMenuAutoConfirm.set_active(False)
			elif widget == cbMenuAutoConfirm:
				if cbMenuConfirmWithClick:
					cbMenuConfirmWithClick.set_active(False)
				if cbMenuAutoCancel:
					cbMenuAutoCancel.set_active(False)
			elif widget == cbMenuAutoCancel:
				if cbMenuAutoConfirm:
					cbMenuAutoConfirm.set_active(False)

	def on_cbMenus_changed(self, *a):
		"""Called when user changes any menu settings"""
		if self._recursing:
			return None
		cbMenuConfirmWithClick = self.builder.get_object("cbMenuConfirmWithClick")
		cbMenuAutoConfirm = self.builder.get_object("cbMenuAutoConfirm")
		cbMenuAutoCancel = self.builder.get_object("cbMenuAutoCancel")
		lblControlWith = self.builder.get_object("lblControlWith")
		cbControlWith = self.builder.get_object("cbControlWith")
		lblConfirmWith = self.builder.get_object("lblConfirmWith")
		cbConfirmWith = self.builder.get_object("cbConfirmWith")
		lblCancelWith = self.builder.get_object("lblCancelWith")
		cbCancelWith = self.builder.get_object("cbCancelWith")

		cbm = self.builder.get_object("cbMenuType")
		menu_type = cbm.get_model().get_value(cbm.get_active_iter(), 1)

		if cbControlWith:
			sensitive = True
			if menu_type == "quickmenu":
				sensitive = False
			lblControlWith.set_sensitive(sensitive)
			cbControlWith.set_sensitive(sensitive)

		if cbConfirmWith:
			sensitive = True
			if cbMenuAutoConfirm and cbMenuAutoConfirm.get_active():
				sensitive = False
			if cbMenuConfirmWithClick and cbMenuConfirmWithClick.get_active():
				sensitive = False
			if menu_type == "quickmenu":
				sensitive = False
			lblConfirmWith.set_sensitive(sensitive)
			cbConfirmWith.set_sensitive(sensitive)

		if cbCancelWith:
			sensitive = True
			if cbMenuAutoCancel and cbMenuAutoCancel.get_active():
				sensitive = False
			if menu_type == "quickmenu":
				sensitive = False
			lblCancelWith.set_sensitive(sensitive)
			cbCancelWith.set_sensitive(sensitive)

		if cbMenuAutoConfirm:
			sensitive = True
			if menu_type == "quickmenu":
				sensitive = False
			cbMenuAutoConfirm.set_sensitive(sensitive)

		name = self.get_selected_menu()
		if name == "":
			return self.on_new_menu_selected()
		if name:
			# There is some menu choosen
			self.builder.get_object("btEditMenu").set_sensitive(name not in MenuEditor.OPEN)
			params = [name]

			cow = SAME
			if cbMenuAutoConfirm and cbMenuAutoConfirm.get_active():
				cow = SAME
			elif cbMenuConfirmWithClick and cbMenuConfirmWithClick.get_active():
				cow = DEFAULT
			elif cbConfirmWith:
				cow = cbConfirmWith.get_model().get_value(cbConfirmWith.get_active_iter(), 1)
				if cow != DEFAULT:
					cow = getattr(SCButtons, cow)

			caw = DEFAULT
			if cbMenuAutoCancel and cbMenuAutoCancel.get_active():
				caw = DEFAULT
			elif cbCancelWith:
				caw = cbCancelWith.get_model().get_value(cbCancelWith.get_active_iter(), 1)
				if caw != DEFAULT:
					caw = getattr(SCButtons, caw)

			params += [self.get_control_with(), cow, caw]

			# Hide / apply and display 'Items per row' selector if it exists in UI
			if self.builder.get_object("rvMenuSize"):
				spMenuSize = self.builder.get_object("spMenuSize")
				menu_type = cbm.get_model().get_value(cbm.get_active_iter(), 1)
				if menu_type == "gridmenu":
					self.update_size_display(GridMenuAction("dummy"))
					size = int(spMenuSize.get_adjustment().get_value())
					if size > 0:
						# size is 2nd parameter
						params += [False, size]
				elif menu_type == "radialmenu":
					self.update_size_display(RadialMenuAction("dummy"))
					size = int(spMenuSize.get_adjustment().get_value())
					if size > 0 and size < 100:
						# Both 0 and 100 means default here
						# size is 2nd parameter
						params += [False, size]
				elif menu_type == "hmenu":
					self.update_size_display(HorizontalMenuAction("dummy"))
					size = int(spMenuSize.get_adjustment().get_value())
					if size > 1:
						# Size 0 and 1 means default here
						# size is 2nd parameter
						params += [False, size]
				else:
					# , "radialmenu"):
					self.update_size_display(None)

			# Grab menu type and choose apropriate action
			action = NoAction()
			if cbm and menu_type == "gridmenu":
				# Grid menu
				action = GridMenuAction(*params)
			elif cbm and menu_type == "radialmenu":
				# Circular menu
				action = RadialMenuAction(*params)
			elif cbm and menu_type == "hmenu":
				# Horizontal menu
				action = HorizontalMenuAction(*params)
			elif cbm and menu_type == "quickmenu":
				# Horizontal menu
				action = QuickMenuAction(name)
			else:
				# Normal menu
				action = MenuAction(*params)

			# Apply Menu Position options, if such block exists in UI
			if self.builder.get_object("spMenuPosX"):
				cbMenuPosX = self.builder.get_object("cbMenuPosX")
				cbMenuPosY = self.builder.get_object("cbMenuPosY")
				x = int(self.builder.get_object("spMenuPosX").get_value())
				y = int(self.builder.get_object("spMenuPosY").get_value())
				x *= cbMenuPosX.get_model().get_value(cbMenuPosX.get_active_iter(), 0)
				y *= cbMenuPosY.get_model().get_value(cbMenuPosY.get_active_iter(), 0)
				if (x, y) != MenuAction.DEFAULT_POSITION:
					action = PositionModifier(x, y, action)

			self.editor.set_action(action)

	def on_new_menu_selected(self, make_copy=False):
		# 'New menu' selected
		self.load_menu_list()
		log.debug("Creating editor for new menu")
		me = MenuEditor(self.app, self.on_menu_changed)
		if make_copy:
			name = self.get_selected_menu()
			log.debug("Copying %s", name)
			me = MenuEditor(self.app, self.on_menu_changed)
			me.set_menu(name)
		me.set_new_menu()
		me.allow_menus(self.allow_globals, self.allow_in_profile)
		me.show(self.editor.window)

	def update_size_display(self, action):
		"""Displays or hides menu size area and upadates text displayed in it.
		Returns True if action is menuaction where changing size makes sense
		"""
		rvMenuSize = self.builder.get_object("rvMenuSize")
		lblMenuSize = self.builder.get_object("lblMenuSize")
		spMenuSize = self.builder.get_object("spMenuSize")
		sclMenuSize = self.builder.get_object("sclMenuSize")
		if isinstance(action, GridMenuAction):
			spMenuSize.set_visible(True)
			sclMenuSize.set_visible(False)
			lblMenuSize.set_text(_("Items per row"))
			rvMenuSize.set_reveal_child(True)
			return True
		if isinstance(action, (RadialMenuAction, HorizontalMenuAction)):
			spMenuSize.set_visible(False)
			sclMenuSize.set_visible(True)
			lblMenuSize.set_text(_("Size"))
			rvMenuSize.set_reveal_child(True)
			return True
		rvMenuSize.set_reveal_child(False)
		return False

	def get_control_with(self):
		"""Returns value of "Control With" combo or STICK if there is none"""
		cbControlWith = self.builder.get_object("cbControlWith")
		if cbControlWith:
			return cbControlWith.get_model().get_value(cbControlWith.get_active_iter(), 1)
		return STICK

	def on_spMenuSize_format_value(self, spinner):
		val = int(spinner.get_adjustment().get_value())
		if val < 1:
			spinner.get_buffer().set_text(_("auto"), -1)
		else:
			spinner.get_buffer().set_text(str(val), -1)
		return True

	def on_sclMenuSize_format_value(self, scale, val):
		cbm = self.builder.get_object("cbMenuType")
		menu_type = cbm.get_model().get_value(cbm.get_active_iter(), 1)
		if menu_type == "radialmenu":
			if val < 1:
				return _("default")
			return "%s%%" % (int(val),)
		# if menu_type == "hmenu"
		val = int(val)
		if val < 2:
			return _("default")
		return str(int(val))



#!/usr/bin/env python3
"""SC-Controller - Controller Widget

Button that user can click to choose emulated action for physical button, axis
or pad.

Wraps around actual button defined in glade file.
"""

import itertools
import logging
import os

from gi.repository import Gdk, Gtk, Pango

from scc.actions import Action, MultiAction
from scc.constants import GYRO, LEFT, RIGHT, STICK, SCButtons
from scc.gui.ae.gyro_action import is_gyro_enable
from scc.modifiers import DoubleclickModifier
from scc.profile import Profile
from scc.tools import _, nameof

log = logging.getLogger("ControllerWidget")

TRIGGERS = ["LT", "RT"]
PADS = [Profile.LPAD, Profile.RPAD, Profile.CPAD]
STICKS = [STICK, Profile.RSTICK, Profile.DPAD]
GYROS = [GYRO]
PRESSABLE = [SCButtons.LPAD, SCButtons.RPAD, SCButtons.STICKPRESS, SCButtons.CPADPRESS]
_NOT_BUTTONS = PADS + STICKS + GYROS + TRIGGERS
_NOT_BUTTONS += [x + "TOUCH" for x in PADS]
BUTTONS = [b for b in SCButtons if b.name not in _NOT_BUTTONS]
LONG_TEXT = 16


class ControllerWidget:
	ACTION_CONTEXT = None

	def __init__(self, app, id, use_icon, widget):
		self.app = app
		self.id = id
		self.name = id if type(id) in (str,) else id.name
		self.widget = widget

		self.label = Gtk.Label()
		self.label.set_ellipsize(Pango.EllipsizeMode.END)
		self.icon = Gtk.Image.new_from_file(self.get_image()) if use_icon else None
		self.update()

		self.widget.connect("enter", self.on_cursor_enter)
		self.widget.connect("leave", self.on_cursor_leave)
		self.widget.connect("clicked", self.on_click)
		self.widget.connect("button-release-event", self.on_button_release)

	def get_image(self):
		return os.path.join(self.app.imagepath, self.name + ".svg")

	def update(self):
		self.label.set_label(_("(no action)"))

	def on_click(self, *a):
		self.app.show_editor(self.id)

	def on_button_release(self, bt, event):
		if event.button == 3:
			# Rightclick
			self.app.show_context_menu(self.id)

	def on_cursor_enter(self, *a):
		self.app.hilight(self.name)

	def on_cursor_leave(self, *a):
		self.app.hilight(None)


class ControllerButton(ControllerWidget):
	ACTION_CONTEXT = Action.AC_BUTTON

	def __init__(self, app, name, use_icon, widget):
		ControllerWidget.__init__(self, app, name, use_icon, widget)

		if use_icon:
			vbox = Gtk.Box(Gtk.Orientation.HORIZONTAL)
			vbox.set_spacing(6)
			separator = Gtk.Separator(orientation=Gtk.Orientation.VERTICAL)
			vbox.pack_start(self.icon, False, False, 1)
			vbox.pack_start(separator, False, False, 1)
			vbox.pack_start(self.label, False, True, 1)
			self.widget.add(vbox)
		else:
			self.widget.add(self.label)
		self.widget.show_all()
		self.label.set_max_width_chars(LONG_TEXT)
		if name == "C":
			self.label.set_max_width_chars(10)

	def update(self):
		if self.id in SCButtons.__members__.values() and self.id in self.app.current.buttons:
			txt = self.app.current.buttons[self.id].describe(self.ACTION_CONTEXT)
			if len(txt) > LONG_TEXT or "\n" in txt:
				txt = "\n".join(txt.split("\n")[0:2])
				txt = txt.replace("<", "&lt;").replace(">", "&gt;")
				self.label.set_markup("<small>%s</small>" % (txt,))
			else:
				txt = txt.replace("<", "&lt;").replace(">", "&gt;")
				self.label.set_markup(txt)
		else:
			self.label.set_label(_("(no action)"))


class ControllerStick(ControllerWidget):
	ACTION_CONTEXT = Action.AC_STICK

	def __init__(self, app, id, use_icon, enable_press, widget):
		self.pressed = Gtk.Label() if enable_press else None
		self.click_button = SCButtons.STICKPRESS if id == STICK else SCButtons.RSTICKPRESS
		ControllerWidget.__init__(self, app, id, use_icon, widget)

		grid = Gtk.Grid()
		grid.set_column_spacing(6)
		self.widget.set_events(Gdk.EventMask.POINTER_MOTION_MASK)
		self.widget.connect("motion-notify-event", self.on_cursor_motion)
		self.label.set_max_width_chars(LONG_TEXT)
		if self.pressed:
			self.label.set_halign(Gtk.Align.START)
			self.pressed.set_halign(Gtk.Align.START)
			self.pressed.set_valign(Gtk.Align.END)
			self.pressed.set_ellipsize(Pango.EllipsizeMode.END)
			self.pressed.set_max_width_chars(LONG_TEXT)
			grid.attach(self.pressed, 2, 2, 1, 1)
		else:
			self.label.set_halign(Gtk.Align.CENTER)
			self.label.set_valign(Gtk.Align.CENTER)
			self.pressed = None
		if self.icon:
			grid.attach(self.icon, 1, 1, 1, 2)
		grid.attach(self.label, 2, 1, 1, 1)
		self.over_icon = False
		self.enable_press = enable_press
		self.widget.add(grid)
		self.widget.show_all()

	def on_cursor_enter(self, *a):
		return

	def on_click(self, *a):
		if self.over_icon and self.enable_press:
			self.app.show_editor(self.click_button)
		else:
			self.app.show_editor(self.id)

	def on_cursor_motion(self, trash, event):
		# self.icon.get_allocation().x + self.icon.get_allocation().width	# yields nonsense
		ix2 = 74
		# Check if cursor is placed on icon
		what = None
		if event.x < ix2:
			what = {
				Profile.LPAD: LEFT,
				Profile.RPAD: RIGHT,
				Profile.CPAD: nameof(SCButtons.CPADPRESS),
				Profile.STICK: nameof(SCButtons.STICKPRESS),
				Profile.RSTICK: nameof(SCButtons.RSTICKPRESS),
				Profile.DPAD: None,
			}.get(self.name)
		if what:
			self.app.hilight(what)
			self.over_icon = True
		else:
			self.app.hilight(self.name)
			self.over_icon = False

	def _set_label(self, action):
		self.label.set_label(action.describe(self.ACTION_CONTEXT))

	def update(self):
		if self.id == Profile.STICK:
			self._set_label(self.app.current.stick)
		elif self.id == Profile.RSTICK:
			self._set_label(self.app.current.rstick)
		elif self.id == Profile.DPAD:
			self._set_label(self.app.current.pads[Profile.DPAD])
		if self.click_button and self.pressed:
			action = self.app.current.buttons[self.click_button]
			self._update_pressed(action)

	def _update_pressed(self, action):
		escape = lambda t: t.replace("<", "&lt;").replace(">", "&gt;")
		if isinstance(action, DoubleclickModifier):
			lines = []
			if action.normalaction:
				txt = action.normalaction.describe(self.ACTION_CONTEXT)
				lines.append("Pressed: %s" % (escape(txt),))
			if action.holdaction:
				txt = action.holdaction.describe(self.ACTION_CONTEXT)
				lines.append("Hold: %s" % (escape(txt),))
			self.pressed.set_markup("<small>%s</small>" % ("\n".join(lines),))
		else:
			txt = escape(action.describe(self.ACTION_CONTEXT))
			self.pressed.set_markup("<small>Pressed: %s</small>" % (txt,))


class ControllerTrigger(ControllerButton):
	ACTION_CONTEXT = Action.AC_TRIGGER

	def update(self):
		# TODO: Use LT and RT in profile as well
		side = LEFT if self.id == "LT" else RIGHT
		if self.id in TRIGGERS and side in self.app.current.triggers:
			self.label.set_label(self.app.current.triggers[side].describe(self.ACTION_CONTEXT))
		else:
			self.label.set_label(_("(no action)"))


class ControllerPad(ControllerStick):
	ACTION_CONTEXT = Action.AC_PAD

	def __init__(self, app, name, use_icon, enable_press, widget):
		ControllerStick.__init__(self, app, name, use_icon, enable_press, widget)
		if name in (Profile.LPAD, Profile.RPAD):
			self.click_button = getattr(SCButtons, name)
		elif name == Profile.CPAD:
			self.click_button = SCButtons.CPADPRESS

	def update(self):
		if self.id == Profile.LPAD:
			action = self.app.current.pads[Profile.LEFT]
			pressed = self.app.current.buttons[SCButtons.LPAD]
		elif self.id == Profile.RPAD:
			action = self.app.current.pads[Profile.RIGHT]
			pressed = self.app.current.buttons[SCButtons.RPAD]
		else:
			action = self.app.current.pads[Profile.CPAD]
			pressed = self.app.current.buttons[SCButtons.CPADPRESS]

		self._set_label(action)
		if self.pressed:
			self._update_pressed(pressed)


class ControllerGyro(ControllerWidget):
	ACTION_CONTEXT = Action.AC_GYRO

	def __init__(self, app, name, use_icon, widget):
		self.pressed = Gtk.Label()
		ControllerWidget.__init__(self, app, name, use_icon, widget)

		grid = Gtk.Grid()
		grid.set_column_spacing(6)
		self.label.set_max_width_chars(LONG_TEXT)
		self.label.set_halign(Gtk.Align.START)
		self.pressed.set_max_width_chars(LONG_TEXT)
		self.pressed.set_halign(Gtk.Align.START)
		self.pressed.set_valign(Gtk.Align.END)
		grid.attach(self.icon, 1, 1, 1, 2)
		grid.attach(self.label, 2, 1, 1, 1)
		grid.attach(self.pressed, 2, 2, 1, 1)
		self.over_icon = False
		self.widget.add(grid)
		self.widget.show_all()

	def on_click(self, *a):
		self.app.show_editor(self.id)

	def _set_label(self, action):
		if is_gyro_enable(action):
			action = next(itertools.islice(action.mods.values(), 0, 1)) or action.default
		if isinstance(action, MultiAction):
			rv = []
			for a in action.actions:
				d = a.describe(self.ACTION_CONTEXT)
				if d not in rv:
					rv.append(d)
			self.label.set_label("\n".join(rv))
			return
		self.label.set_label(action.describe(self.ACTION_CONTEXT))

	def update(self):
		self._set_label(self.app.current.gyro)



"""SC-Controller - Device Monitor.

Extends eudevmonitor with options to register callbacks and
manage plugging/releasing devices.
"""

from __future__ import annotations

from ctypes.util import find_library
from typing import TYPE_CHECKING

from ioctl_opt import IOR

from scc.lib.eudevmonitor import Eudev, Monitor

if TYPE_CHECKING:
	from scc.sccdaemon import SCCDaemon
import ctypes
import fcntl
import logging
import os
import re

log = logging.getLogger("DevMon")

RE_BT_NUMBERS = re.compile(r"[0-9A-F]{4}:([0-9A-F]{4}):([0-9A-F]{4}).*")
RE_HCI_LINK = re.compile(r"hci([0-9]+):([0-9]+)")
HCIGETCONNLIST = IOR(ord("H"), 212, ctypes.c_int)
HAVE_BLUETOOTH_LIB = False
try:
	btlib_name = find_library("bluetooth")
	assert btlib_name
	btlib = ctypes.CDLL(btlib_name)
	HAVE_BLUETOOTH_LIB = True
except Exception:
	pass


def _disconnect_bluez(device_path: str) -> None:
	# Import lazily so importing DeviceMonitor does not require PyGObject.
	from gi.repository import Gio

	bus = Gio.bus_get_sync(Gio.BusType.SYSTEM, None)
	bus.call_sync(
		"org.bluez",
		device_path,
		"org.bluez.Device1",
		"Disconnect",
		None,
		None,
		Gio.DBusCallFlags.NONE,
		5000,
		None,
	)


class DeviceMonitor(Monitor):
	BT_DISCOVERY_RETRY_DELAY = 0.25
	BT_DISCOVERY_RETRY_COUNT = 20

	def __init__(self, *a) -> None:
		Monitor.__init__(self, *a)
		self.daemon: SCCDaemon | None = None
		self.dev_added_cbs = {}
		self.dev_removed_cbs = {}
		self.bt_addresses = {}
		self.known_devs = {}
		self._pending_bt = {}

	# removed_cb type can be None
	def add_callback(self, subsystem: str, vendor_id: int, product_id: int, added_cb, removed_cb) -> None:
		"""Add a function that is called when eudev monitor detects new, ready to use device.

		This has to be called from something called by init_drivers method.
		"""
		key = (subsystem, vendor_id, product_id)
		# TODO type this, might need 3.11 for https://docs.python.org/3/library/typing.html#typing.Self ?
		# print(f"add_callback for subsystem {subsystem} for {vendor_id}:{product_id} - existing {self.dev_added_cbs}; of type {type(self.dev_added_cbs)} with existing removals {self.dev_removed_cbs}; of type {type(self.dev_removed_cbs)}\n")
		# print(f"  {added_cb}; of type {type(added_cb)} with additional removals {added_cb}; of type {type(added_cb)}\n")
		assert key not in self.dev_added_cbs, (
			f"DeviceMonitor add_callback found an already present device: \n    {key[0], key[1], key[2]} in hex: {hex(key[1]), hex(key[2])}\n  in existing DeviceMonitor cbs:\n    {self.dev_added_cbs}\n \
			Perhaps a custom device was added to config that we already considered earlier? Check $config_dir/scc/devices/"
		)
		self.match_subsystem(subsystem)

		self.dev_added_cbs[key] = added_cb
		self.dev_removed_cbs[key] = removed_cb

	def add_remove_callback(self, syspath: str, cb) -> None:
		"""Add (possibly replace) callback that will be called once device with specified syspath is disconnected."""
		if syspath in self.known_devs:
			vendor, product, old_cb = self.known_devs.pop(syspath)
			self.known_devs[syspath] = (vendor, product, cb)

	def start(self) -> None:
		"""Register poller and starts listening for events."""
		if not HAVE_BLUETOOTH_LIB:
			log.warning("Failed to load libbluetooth.so, bluetooth support will be incomplete")
		poller = self.daemon.poller
		poller.register(self.fileno(), poller.POLLIN, self.on_data_ready)
		Monitor.start(self)

	def _on_new_syspath(self, subsystem: str, syspath: str) -> None:
		# Bluetooth adapters are reported as e.g. hci0, while actual device
		# links are reported as hci0:<connection handle>. Only links can be
		# resolved to a HID device and vendor/product pair.
		if subsystem == "bluetooth" and ":" not in os.path.basename(syspath):
			return
		try:
			if subsystem == "input":
				vendor, product = None, None
			else:
				vendor, product = self.get_vendor_product(syspath, subsystem)
		except OSError:
			# Cannot grab vendor & product, probably subdevice or bus itself

			# BlueZ creates the bluetooth link before UHID has created the HID
			# hierarchy from which vendor and product IDs are read. Retry without
			# blocking the daemon while that hierarchy is being constructed.
			if subsystem == "bluetooth" and os.path.exists(syspath):
				log.debug(f"Failed to get VID:PID for {syspath}, will retry…")
				self._schedule_bt_retry(syspath)
			return
		key = (subsystem, vendor, product)
		cb = self.dev_added_cbs.get(key)
		rem_cb = self.dev_removed_cbs.get(key)
		if cb:
			self.known_devs[syspath] = (vendor, product, rem_cb)
			try:
				if cb(syspath, vendor, product) is None:
					del self.known_devs[syspath]
					if subsystem == "bluetooth" and os.path.exists(syspath):
						self._schedule_bt_retry(syspath)
					return
			except Exception as e:
				log.exception(e)
				del self.known_devs[syspath]
		self._pending_bt.pop(syspath, None)

	def _schedule_bt_retry(self, syspath: str) -> None:
		attempt, _task = self._pending_bt.get(syspath, (0, None))
		if attempt >= self.BT_DISCOVERY_RETRY_COUNT:
			log.warning("Bluetooth device did not become ready: %s", syspath)
			self._pending_bt.pop(syspath, None)
			return

		def retry() -> None:
			if os.path.exists(syspath) and syspath not in self.known_devs:
				self._get_hci_addresses()
				self._on_new_syspath("bluetooth", syspath)

		task = self.daemon.get_scheduler().schedule(self.BT_DISCOVERY_RETRY_DELAY, retry)
		self._pending_bt[syspath] = (attempt + 1, task)

	def _cancel_bt_retry(self, syspath: str) -> None:
		pending = self._pending_bt.pop(syspath, None)
		if pending:
			_attempt, task = pending
			task.cancel()

	def _get_hci_addresses(self) -> None:
		if not HAVE_BLUETOOTH_LIB:
			return
		cl = hci_conn_list_req()
		cl.dev_id = btlib.hci_get_route(ctypes.c_void_p(0))
		if cl.dev_id < 0 or cl.dev_id > 65534:
			return
		cl.conn_num = 256

		s = btlib.hci_open_dev(cl.dev_id)
		if fcntl.ioctl(s, HCIGETCONNLIST, cl, True):
			log.error("Failed to list bluetooth collections")
			return

		for i in range(cl.conn_num):
			ci = cl.conn_info[i]
			id = f"hci{cl.dev_id}:{ci.handle}"
			address = ":".join([hex(x).lstrip("0x").zfill(2).upper() for x in reversed(ci.bdaddr)])
			self.bt_addresses[id] = address

	def disconnect_bluetooth(self, syspath: str) -> None:
		"""Ask BlueZ to disconnect the device represented by an hciX:handle syspath."""
		match = RE_HCI_LINK.fullmatch(os.path.basename(syspath))
		if match is None:
			raise OSError(f"Cannot identify Bluetooth connection from {syspath}")

		dev_id, _handle = match.groups()
		link_name = os.path.basename(syspath)
		address = self.bt_addresses.get(link_name)
		if address is None:
			self._get_hci_addresses()
			address = self.bt_addresses.get(link_name)
		if address is None:
			raise OSError(f"Cannot determine Bluetooth address for {link_name}")

		device_path = "/org/bluez/hci{}/dev_{}".format(dev_id, address.replace(":", "_"))
		_disconnect_bluez(device_path)

	def _dev_for_hci(self, sys_bus_path: str) -> str | None:
		"""For given syspath leading to ../hciX:ABCD, returns input device node.

		For example "a0:5a:5d:87:82:17" or "vial:f64c2b3c"
		"""
		name = sys_bus_path.rsplit("/", maxsplit=1)[-1]
		if ":" not in name:
			return None
		addr = self.bt_addresses.get(name)
		for fname in os.listdir("/sys/bus/hid/devices/"):
			node = os.path.join("/sys/bus/hid/devices/", fname)
			try:
				node_addr = DeviceMonitor._find_bt_address(node)
				# Joe: somehow my node_addr is in lowercase
				if node_addr is not None:
					node_addr = str.upper(node_addr)
			except OSError:
				continue
			try:
				if node_addr is not None and addr is not None:
					# SteamOS 3 "Holo" return caps
					if node_addr.lower() == addr.lower():
						return node
			# None
			except AttributeError:
				pass
		return None

	def on_data_ready(self, *a) -> None:
		event = self.receive_device()
		if event:
			if event.action == "bind" and event.initialized:
				if event.syspath not in self.known_devs:
					self._on_new_syspath(event.subsystem, event.syspath)
			elif event.action == "add" and event.initialized and event.subsystem in ("input", "bluetooth"):
				# those are not bound
				if event.syspath not in self.known_devs:
					if event.subsystem == "bluetooth":
						self._get_hci_addresses()
					self._on_new_syspath(event.subsystem, event.syspath)
			elif event.action in ("remove", "unbind"):
				self._cancel_bt_retry(event.syspath)
				if event.syspath in self.known_devs:
					vendor, product, cb = self.known_devs.pop(event.syspath)
					if cb:
						cb(event.syspath, vendor, product)

	def rescan(self) -> None:
		"""Scan and call callbacks for already connected devices."""
		self._get_hci_addresses()
		enumerator = self._eudev.enumerate()
		subsystem_to_vp_to_callback = {}

		for key, cb in self.dev_added_cbs.items():
			subsystem, vendor_id, product_id = key
			enumerator.match_subsystem(subsystem)
			if subsystem not in subsystem_to_vp_to_callback:
				subsystem_to_vp_to_callback[subsystem] = {}
			subsystem_to_vp_to_callback[subsystem][vendor_id, product_id] = cb

		for syspath in enumerator:
			if syspath not in self.known_devs:
				try:
					subsystem = DeviceMonitor.get_subsystem(syspath)
				except OSError:
					continue
				if subsystem in subsystem_to_vp_to_callback:
					self._on_new_syspath(subsystem, syspath)

	def get_vendor_product(self, syspath: str, subsystem: str | None = None) -> tuple[int, int]:
		"""For given syspath, reads and returns (vendor_id, product_id) as ints.

		May throw all kinds of OSErrors
		"""
		if os.path.exists(os.path.join(syspath, "idVendor")):
			vendor = int(open(os.path.join(syspath, "idVendor")).read().strip(), 16)
			product = int(open(os.path.join(syspath, "idProduct")).read().strip(), 16)
			return vendor, product
		if subsystem is None:
			subsystem = DeviceMonitor.get_subsystem(syspath)
		if subsystem == "bluetooth":
			# Search for folder that matches regular expression...
			names = [name for name in os.listdir(syspath) if os.path.isdir(syspath) and RE_BT_NUMBERS.match(name)]
			if len(names) > 0:
				vendor, product = [int(x, 16) for x in RE_BT_NUMBERS.match(names[0]).groups()]
				return vendor, product
			# Above method works for anything _but_ SteamController
			# For that one, following desperate mess is needed

			node = self._dev_for_hci(syspath)
			if node:
				name = node.split("/")[-1]
				if RE_BT_NUMBERS.match(name):
					vendor, product = [int(x, 16) for x in RE_BT_NUMBERS.match(name).groups()]
					return vendor, product
		raise OSError("Cannot determine vendor and product IDs")

	def get_hidraw(self, syspath: str) -> str | None:
		"""For given syspath, returns name of assotiated hidraw device.

		Returns None if there is no such thing.
		"""
		node = self._dev_for_hci(syspath)
		if node is None:
			return None
		hidrawsubdir = os.path.join(node, "hidraw")
		for fname in os.listdir(hidrawsubdir):
			if fname.startswith("hidraw"):
				return fname
		return None

	@staticmethod
	def _find_bt_address(syspath: str) -> str | None:
		"""Recursivelly searchs for "input*" subdirectories until "uniq" file is found. Then, returns address from that file."""
		uniq = os.path.join(syspath, "uniq")
		if os.path.exists(uniq):
			with open(uniq) as file:
				return file.read().strip()
		for name in os.listdir(syspath):
			if name.startswith("input"):
				path = os.path.join(syspath, name)
				if os.path.isdir(path) and not os.path.islink(path):
					addr = DeviceMonitor._find_bt_address(path)
					if addr:
						return addr
		return None

	@staticmethod
	def get_usb_address(syspath: str) -> tuple[int, int]:
		"""For given syspath, read and return (busnum, devnum) as ints.

		May throw all kinds of OSErrors
		"""
		busnum = int(open(os.path.join(syspath, "busnum")).read().strip())
		devnum = int(open(os.path.join(syspath, "devnum")).read().strip())
		return busnum, devnum

	@staticmethod
	def get_subsystem(syspath: str) -> str:
		"""For given syspath, reads and returns subsystem as string.

		May throw OSError if directory is not readable.
		"""
		return os.readlink(os.path.join(syspath, "subsystem")).split("/")[-1]


class hci_conn_info(ctypes.Structure):
	_fields_ = [
		("handle", ctypes.c_uint16),
		("bdaddr", ctypes.c_uint8 * 6),
		("type", ctypes.c_uint8),
		("out", ctypes.c_uint8),
		("state", ctypes.c_uint16),
		("link_mode", ctypes.c_uint32),
	]


class hci_conn_list_req(ctypes.Structure):
	_fields_ = [
		("dev_id", ctypes.c_uint16),
		("conn_num", ctypes.c_uint16),
		("conn_info", hci_conn_info * 256),
	]


def create_device_monitor(daemon: SCCDaemon) -> DeviceMonitor:
	mon = Eudev().monitor(subclass=DeviceMonitor)
	assert type(mon) is DeviceMonitor  # Satisfy type checker
	mon.daemon = daemon
	return mon



#!/usr/bin/env python3
"""SC-Controller - ProfileSwitcher

Set of widgets designed to allow user to select profile, placed in one Gtk.Box:
 [ Icon | Combo box with profile selection       (ch) | (S) ]

... where (S) is Save button that can be shown on demand and (ch) is change
indicator drawn in combobox.
"""

import logging
import os
import random

from gi.repository import Gio, GLib, GObject, Gtk

from scc.gui.userdata_manager import UserDataManager
from scc.paths import get_controller_icons_path, get_default_controller_icons_path
from scc.tools import _, find_controller_icon, find_profile

log = logging.getLogger("PS")


class ProfileSwitcher(Gtk.EventBox, UserDataManager):
	"""List of signals:
	changed (name, giofile)
		Emited when value of selection combobox is changed
	new-clicked (name)
		Emited when user selects 'new profile' option.
		'name' is of currently selected profile.
	right-clicked ()
		Emited whenm user right-clicks anything
	save-clicked ()
		Emited when user clicks on save button
	switch-to-clicked ()
		Emited when user clicks on "switch to this controller" button
	unknown-profile (name)
		Emited when daemon reports unknown profile for controller.
		'name' is name of reported profile.
	"""

	__gsignals__ = {
		"changed": (GObject.SignalFlags.RUN_FIRST, None, (object, object)),
		"new-clicked": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
		"right-clicked": (GObject.SignalFlags.RUN_FIRST, None, ()),
		"save-clicked": (GObject.SignalFlags.RUN_FIRST, None, ()),
		"switch-to-clicked": (GObject.SignalFlags.RUN_FIRST, None, ()),
		"unknown-profile": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
	}

	SEND_TIMEOUT = 100  # How many ms should switcher wait before sending event
	# about profile being switched

	def __init__(self, imagepath, config):
		Gtk.EventBox.__init__(self)
		UserDataManager.__init__(self)
		self.imagepath = imagepath
		self.config = config
		self._allow_new = False
		self._first_time = True
		self._current = None
		self._recursing = False
		self._timer = None  # Used to prevent sending too many request
		# when user scrolls through combobox
		self._signal = None
		self._controller = None
		self.setup_widgets()

	def setup_widgets(self):
		# Create
		self._icon = Gtk.Image()
		self._model = Gtk.ListStore(str, object, str)
		self._combo = Gtk.ComboBox.new_with_model(self._model)
		self._box = Gtk.Box(Gtk.Orientation.HORIZONTAL, 0)
		self._savebutton = None
		self._switch_to_button = None

		# Setup
		rend1 = Gtk.CellRendererText()
		rend2 = Gtk.CellRendererText()
		self._box.set_spacing(12)
		self._combo.pack_start(rend1, True)
		self._combo.pack_start(rend2, False)
		self._combo.add_attribute(rend1, "text", 0)
		self._combo.add_attribute(rend2, "text", 2)
		self._combo.set_row_separator_func(
			lambda model, iter: model.get_value(iter, 1) is None and model.get_value(iter, 0) == "-",
		)
		self.update_icon()

		# Signals
		self._combo.connect("changed", self.on_combo_changed)
		self.connect("button_press_event", self.on_button_press)

		# Pack
		self._box.pack_start(self._icon, False, True, 0)
		self._box.pack_start(self._combo, True, True, 0)
		self.add(self._box)

	def set_profile(self, name, create=False):
		"""Selects specified profile in UI.
		Returns True on success or False if profile is not in combobox.

		If 'create' is set to True, creates new combobox item if needed.
		"""
		if name is None:
			return None

		if name.endswith(".mod"):
			name = name[0:-4]
		if name.endswith(".sccprofile"):
			name = name[0:-11]
		if "/" in name:
			name = os.path.split(name)[-1]

		self._current = name
		active = self._combo.get_active_iter()
		giofile = None
		for row in self._model:
			if self._model.get_value(row.iter, 1) is not None:
				if name == self._model.get_value(row.iter, 0):
					giofile = self._model.get_value(row.iter, 1)
					if active == row.iter:
						# Already selected
						break
					self._combo.set_active_iter(row.iter)
					break
		if giofile is None and create:
			path = find_profile(name)
			if path:
				giofile = Gio.File.new_for_path(path)
				self._model.insert(0, (name, giofile, None))
				self._combo.set_active(0)

		return giofile != None

	def set_allow_new(self, allow):
		"""Enables or disables creating new profile from this ProfileSwitcher.
		Should be called before set_profile_list.
		"""
		self._allow_new = allow

	def set_allow_switch(self, allow):
		"""Enables or disables profile switching for this ProfileSwitcher.
		When disabled, only save button is be usable.
		"""
		self._combo.set_sensitive(allow)

	def set_profile_list(self, lst):
		"""Fills combobox with given list of available profiles.
		'lst' is expected to be iterable of GIO.File's.
		"""
		self._model.clear()
		i, current_index = 0, 0
		for f in sorted(lst, key=lambda f: f.get_basename()):
			name = f.get_basename()
			if type(name) is not str:
				name = name.decode("utf-8")
			if name.endswith(".mod"):
				continue
			if name.startswith("."):
				continue
			if name.endswith(".sccprofile"):
				name = name[0:-11]
			if name == self._current:
				current_index = i
			self._model.append((name, f, None))
			i += 1
		if self._allow_new:
			self._model.append(("-", None, None))
			self._model.append((_("New profile..."), None, None))

		if self._combo.get_active_iter() is None:
			self._combo.set_active(current_index)

	def get_profile_list(self):
		"""Returns profiles in combobox as iterable of Gio.File's"""
		return (x[1] for x in self._model if x[1] is not None)

	def get_profile_name(self):
		"""Returns name of currently selected profile"""
		return self._model.get_value(self._combo.get_active_iter(), 0)

	def refresh_profile_path(self, name):
		"""Called from main window after profile file is deleted.
		May either change path to profile in default_profiles directory,
		or remove entry entirely.
		"""
		prev = None
		new_path = find_profile(name)
		# Find row with named profile
		for row in self._model:
			if row[0] == name:
				active = self._combo.get_active_iter()
				if new_path is None:
					# Profile was completly removed
					if self._model.get_value(active, 0) == name:
						# If removed profile happends to be active one (what's
						# almost always), switch to previous profile in list
						self._model.remove(row.iter)
						if prev is None:
							# ... unless removed profile was 1st in list. Switch
							# to next in that case
							self._combo.set_active_iter(self._model[0].iter)
						else:
							self._combo.set_active_iter(prev.iter)
					else:
						self._model.remove(row.iter)
				else:
					giofile = Gio.File.new_for_path(new_path)
					self._model.set_value(row.iter, 1, giofile)
					if self._model.get_value(active, 0) == name:
						# Active profile was changed
						self.emit("changed", name, giofile)
				return
			prev = row

	def on_combo_changed(self, cb):
		if self._recursing:
			return

		def run_later():
			name = self._model.get_value(cb.get_active_iter(), 0)
			giofile = self._model.get_value(cb.get_active_iter(), 1)
			GLib.source_remove(self._timer)
			self._timer = None

			if giofile is None:
				# 'New profile selected'
				self._recursing = True
				if self._current is None:
					cb.set_active(0)
				else:
					self.set_profile(self._current)
				self._recursing = False

				self.emit("new-clicked", self.get_profile_name())
			else:
				self._current = name
				self.emit("changed", name, giofile)

		if self._timer is not None:
			GLib.source_remove(self._timer)
		self._timer = GLib.timeout_add(ProfileSwitcher.SEND_TIMEOUT, run_later)

	def on_button_press(self, trash, event):
		if event.button == 3:
			self.emit("right-clicked")

	def on_savebutton_clicked(self, *a):
		self.emit("save-clicked")

	def on_switch_to_clicked(self, *a):
		self.emit("switch-to-clicked")

	def on_daemon_dead(self, *a):
		"""Called from App when connection to daemon is lost"""
		self._first_time = True

	def on_profile_changed(self, c, profile):
		"""Called when controller profile is changed from daemon"""
		if not self.set_profile(profile, True):
			if self._first_time:

				def later():
					# Cannot be executed right away, as profile-changed is
					# emitted before DaemonManager finishes initiaalisation
					self.emit("unknown-profile", profile)

				GLib.idle_add(later)
		self._first_time = False

	def set_profile_modified(self, has_changes, is_template=False):
		"""Called to signalize that profile has changes to save in UI
		by displaying "changed" next to profile name and showing Save button.

		Returns giofile for currently selected profile. If profile is set as
		changed, giofile is automatically changed to 'original/filename.mod',
		so application can save changes without overwriting original wile.
		"""
		if has_changes:
			if not self._savebutton:
				# Save button has to be created
				self._savebutton = ButtonInRevealer("gtk-save", _("Save changes"), self.on_savebutton_clicked)
				self._box.pack_start(self._savebutton, False, True, 0)
				self.show_all()
			self._savebutton.set_reveal_child(True)
			iter = self._combo.get_active_iter()
			if is_template:
				self._model.set_value(iter, 2, _("(changed template)"))
			else:
				self._model.set_value(iter, 2, _("(changed)"))
		else:
			if self._savebutton:
				# Nothing to hide if there is no revealer
				self._savebutton.set_reveal_child(False)
			for i in self._model:
				i[2] = None
			if is_template:
				iter = self._combo.get_active_iter()
				self._model.set_value(iter, 2, _("(template)"))

	def set_switch_to_enabled(self, enabled):
		"""Shows or hides 'switch-to' button
		"""
		if enabled:
			if not self._switch_to_button:
				# Save button has to be created
				self._switch_to_button = ButtonInRevealer(
					"gtk-edit", _("Edit mappings of this controller"), self.on_switch_to_clicked,
				)
				self._box.pack_start(self._switch_to_button, False, True, 0)
				self.show_all()
			self._switch_to_button.set_reveal_child(True)
		elif self._switch_to_button:
			# Nothing to hide if there is no revealer
			self._switch_to_button.set_reveal_child(False)

	def get_file(self):
		"""Returns set profile as GIO file or None if there is no any"""
		return

	def set_controller(self, c):
		if self._signal:
			self._controller.disconnect(self._signal)
			self._signal = None
		self._controller = c
		if c:
			name = self.config.get_controller_config(c.get_id())["name"]
			self._icon.set_tooltip_text(name)
			self._signal = c.connect("profile-changed", self.on_profile_changed)
		else:
			self._icon.set_tooltip_text(_("Profile"))
		self.update_icon()

	def get_controller(self):
		"""Returns controller set by set_controller function"""
		return self._controller

	def update_icon(self):
		"""Changes displayed icon to whatever is currently set in config"""
		# Called internally and from ControllerSettings
		if not self._controller:
			self._icon.set_from_file(os.path.join(self.imagepath, "controller-icon.svg"))
			return

		id = self._controller.get_id()
		cfg = self.config.get_controller_config(id)
		if cfg["icon"]:
			icon = find_controller_icon(cfg["icon"])
			self._icon.set_from_file(icon)
		else:
			log.debug("There is no icon for controller %s, auto assinging one", id)
			paths = [get_default_controller_icons_path(), get_controller_icons_path()]

			def cb(icons):
				if id != self._controller.get_id():
					# Controller was changed before callback was called
					return
				icon = None
				used_icons = {
					self.config["controllers"][x]["icon"]
					for x in self.config["controllers"]
					if "icon" in self.config["controllers"][x]
				}
				tp = "%s-" % (self._controller.get_type(),)
				icons = sorted(os.path.split(x.get_path())[-1] for x in icons)
				log.debug("Searching for icon type: %s", tp.strip("-"))
				for i in icons:
					if i not in used_icons and i.startswith(tp):
						# Unused icon found
						icon = i
						break
				else:
					# All icons are already used, assign anything
					icon = random.choice(icons)
				log.debug("Auto-assigned icon %s for controller %s", icon, id)
				cfg = self.config.get_controller_config(id)
				cfg["icon"] = icon
				self.config.save()
				GLib.idle_add(self.update_icon)

			self.load_user_data(paths, "*.svg", None, cb)


class ButtonInRevealer(Gtk.Revealer):
	def __init__(self, button_name, tooltip, callback):
		Gtk.Revealer.__init__(self)
		self.button = Gtk.Button.new_from_icon_name(button_name, Gtk.IconSize.SMALL_TOOLBAR)
		self.button.connect("clicked", callback)
		self.button.set_tooltip_text(tooltip)
		self.set_reveal_child(False)
		self.set_transition_type(Gtk.RevealerTransitionType.SLIDE_LEFT)
		self.add(self.button)



"""SC-Controller - OSD Menu

Display menu that user can navigate through
"""
from __future__ import annotations

import json
import logging
import os
import sys
from math import atan2, cos, sin
from math import pi as PI
from typing import TYPE_CHECKING

import cairo

from scc.config import Config
from scc.constants import STICK, STICK_PAD_MAX
from scc.gui.svg_widget import SVGEditor, SVGWidget
from scc.lib import xwrappers as X
from scc.menu_data import Separator, Submenu
from scc.osd import OSDWindow
from scc.osd.menu import Menu, MenuIcon
from scc.paths import get_share_path
from scc.tools import degdiff, find_icon

if TYPE_CHECKING:
	from gi.repository import Gtk

log = logging.getLogger("osd.menu")


class RadialMenu(Menu):
	RECOLOR_BACKGROUNDS = ("background", "menuitem_hilight_border", "text")
	RECOLOR_STROKES = ("border", "menuitem_border")
	MIN_DISTANCE = 3000  # Minimal cursor distance from center (in px^2)
	ICON_SIZE = 96

	def __init__(self) -> None:
		Menu.__init__(self, "osd-radial-menu")
		self.angle = 0
		self.rotation = 0
		self.scale = 1.0
		self.items_with_icon = []
		if self.xdisplay is None:
			# Compositors normally provide an RGBA visual, but the API can
			# return null if it is not supported, so retain a rectangular fallback.
			visual = self.get_screen().get_rgba_visual()
			if visual is not None:
				self.set_visual(visual)
				self.set_app_paintable(True)
				self.connect("draw", self._on_draw_clip_circle)

	def create_parent(self) -> SVGWidget:
		background = os.path.join(get_share_path(), "images", "radial-menu.svg")
		self.b = SVGWidget(background)
		self.b.connect("size-allocate", self.on_size_allocate)
		self.recolor()
		return self.b

	def recolor(self):
		config = Config()
		source_colors = {}
		try:
			# Try to read json file and bail out if it fails
			desc = os.path.join(get_share_path(), "images", "radial-menu.svg.json")
			source_colors = json.loads(open(desc).read())["colors"]
		except Exception as e:
			log.warning("Failed to load keyboard description")
			log.warning(e)
			return
		editor = self.b.edit()

		for k in RadialMenu.RECOLOR_BACKGROUNDS:
			if k in config["osd_colors"] and k in source_colors:
				editor.recolor_background(source_colors[k], config["osd_colors"][k])
		editor.recolor_background(source_colors["background"], config["osd_colors"]["background"])

		for k in RadialMenu.RECOLOR_STROKES:
			if k in config["osd_colors"] and k in source_colors:
				print("REC", source_colors[k], config["osd_colors"][k])
				editor.recolor_strokes(source_colors[k], config["osd_colors"][k])

		editor.commit()

	def on_size_allocate(self, trash, allocation):
		"""(Re)centers all icons when menu is displayed or size is changed"""
		cx = allocation.width * self.scale * 0.5
		cy = allocation.height * self.scale * 0.5
		radius = min(cx, cy) * 2 / 3
		for i in self.items_with_icon:
			angle, icon = float(i.a) * PI / 180.0, i.icon_widget
			x, y = cx + sin(angle) * radius, cy - cos(angle) * radius
			x = x - (self.ICON_SIZE * self.scale * 0.5)
			y = y - (self.ICON_SIZE * self.scale * 0.5)
			i.icon_widget.get_parent().move(i.icon_widget, x, y)

	def get_window_size(self):
		w, h = Menu.get_window_size(self)
		if self.scale != 1.0:
			w = int(w * self.scale)
			h = int(h * self.scale)
		return w, h

	def _add_arguments(self):
		Menu._add_arguments(self)
		self.argparser.add_argument("--rotation", type=float, default=0, help="rotates input by angle (default: 0)")

	def parse_argumets(self, argv):
		self.editor = self.b.edit()
		rv = Menu.parse_argumets(self, argv)
		self.rotation = self.args.rotation
		if rv:
			self.enable_cursor()
		return rv

	def generate_widget(self, item):
		if isinstance(item, (Separator, Submenu)) or item.id is None:
			# Labels and separators, radial menu can't show these
			return None
		e = self.editor.clone_element("menuitem_template")
		SVGEditor.set_text(e, item.label)
		e.attrib["id"] = "menuitem_" + item.id
		return e

	def pack_items(self, trash, items):
		if self._size > 0 and self._size < 100:
			self.scale = self._size / 100.0
			root = SVGEditor.get_element(self.editor, "root")
			SVGEditor.scale(root, self.scale)
		pb = self.b.get_pixbuf()
		# Image width is not scaled as everything bellow operates
		# in 'root' object coordinate space
		image_width = pb.get_width()

		index = 0
		item_offset = 360.0 / len(self.items)
		a1 = (-90.0 - item_offset * 0.5) * PI / 180.0
		a2 = (-90.0 + item_offset * 0.5) * PI / 180.0
		for i in self.items_with_icon:
			i.icon_widget.get_parent().remove_child(i.icon_widget)
		self.items_with_icon = []
		for i in items:
			# Set size of each arc
			if SVGEditor.get_element(i.widget, "arc") is not None:
				l = SVGEditor.get_element(i.widget, "arc")
				radius = float(l.attrib["radius"])  # TODO: Find how to get value of 'sodipodi:rx'
				l.attrib["d"] = l.attrib["d-template"] % (
					radius * cos(a1) + image_width / 2,
					radius * sin(a1) + image_width / 2,
					radius * cos(a2) + image_width / 2,
					radius * sin(a2) + image_width / 2,
				)
			# Rotate arc to correct position
			i.a = (360.0 / float(len(self.items))) * float(index)
			SVGEditor.rotate(i.widget, i.a, image_width * 0.5, image_width * 0.5)
			# Check if there is any icon
			icon_file, has_colors = find_icon(i.icon, False) if hasattr(i, "icon") else (None, False)
			if icon_file:
				# Icon - hide all text and place MenuIcon widget on top of image
				self.editor.remove_element(SVGEditor.get_element(i.widget, "menuitem_text"))
				self.editor.remove_element(SVGEditor.get_element(i.widget, "line0"))
				self.editor.remove_element(SVGEditor.get_element(i.widget, "line2"))
				i.icon_widget = MenuIcon(icon_file, has_colors)
				i.icon_widget.set_name("osd-radial-menu-icon")
				i.icon_widget.set_size_request(self.ICON_SIZE * self.scale, self.ICON_SIZE * self.scale)
				self.b.get_parent().put(i.icon_widget, 200, 200)
				self.items_with_icon.append(i)
			else:
				# No icon - rotate text in arc to other direction to keep it horisontal
				if SVGEditor.get_element(i.widget, "menuitem_text") is not None:
					l = SVGEditor.get_element(i.widget, "menuitem_text")
					l.attrib["id"] = "text_" + i.id
					l.attrib["transform"] = "%s rotate(%s)" % (l.attrib["transform"], -i.a)
				# Place up to 3 lines of item label
				label = i.label.split("\n")
				first_line = 0
				if len(label) == 1:
					self.editor.remove_element(SVGEditor.get_element(i.widget, "line0"))
					self.editor.remove_element(SVGEditor.get_element(i.widget, "line2"))
					first_line = 1
				elif len(label) == 2:
					self.editor.remove_element(SVGEditor.get_element(i.widget, "line0"))
					first_line = 1
				for line in range(len(label)):
					l = SVGEditor.get_element(i.widget, "line%s" % (first_line + line,))
					if l is None:
						break
					SVGEditor.set_text(l, label[line])
			# Continue with next menu item
			i.index = index

			index += 1

		self.editor.remove_element("menuitem_template")
		self.editor.commit()
		del self.editor

	def _on_draw_clip_circle(self, widget: Gtk.Widget, cr: cairo.Context[cairo.Surface]) -> bool:
		"""Wayland circular shaping

		Clear the surface to transparent and clip all drawing (background, arcs, icons)
		to a centered circle inscribed in the window.

		The transparent corners give the same round appearance the X SHAPE extension produces on X11.

		Returns False so the default handler then draws the children within this clip.
		"""
		cr.set_source_rgba(0, 0, 0, 0)
		cr.set_operator(cairo.OPERATOR_SOURCE)
		cr.paint()
		cr.set_operator(cairo.OPERATOR_OVER)
		w = widget.get_allocated_width()
		h = widget.get_allocated_height()
		cr.arc(w / 2.0, h / 2.0, min(w, h) / 2.0, 0, 2 * PI)
		cr.clip()
		return False

	def show(self) -> None:
		OSDWindow.show(self)

		pb = self.b.get_pixbuf()

		width = int(pb.get_width() * self.scale * self.get_scale_factor())
		height = int(pb.get_height() * self.scale * self.get_scale_factor())
		self.f.move(self.cursor, int(width / 2), int(height / 2))

		if self.xdisplay is None:
			# Clipping to circle on Wayland is handled in __init__()
			return

		win = X.XID(self.get_window().get_xid())
		pixmap = X.create_pixmap(self.xdisplay, win, width, height, 1)

		gc = X.create_gc(self.xdisplay, pixmap, 0, None)
		X.set_foreground(self.xdisplay, gc, 0)
		X.fill_rectangle(self.xdisplay, pixmap, gc, 0, 0, width, height)
		X.set_foreground(self.xdisplay, gc, 1)
		X.set_background(self.xdisplay, gc, 1)

		r = int(width * 0.985)
		x = int((width - r) / 2)

		X.fill_arc(self.xdisplay, pixmap, gc, x, x, r, r, 0, 360 * 64)

		X.flush_gc(self.xdisplay, gc)
		X.flush(self.xdisplay)

		X.shape_combine_mask(self.xdisplay, win, X.SHAPE_BOUNDING, 0, 0, pixmap, X.SHAPE_SET)

		X.flush(self.xdisplay)

	def select(self, i) -> None:
		if type(i) == int:
			i = self.items[i]
		if self._selected and hasattr(self._selected, "icon_widget"):
			if self._selected.icon_widget:
				self._selected.icon_widget.set_name("osd-radial-menu-icon")
		self._selected = i
		if hasattr(self._selected, "icon_widget") and self._selected.icon_widget:
			self._selected.icon_widget.set_name("osd-radial-menu-icon-selected")
		self.b.hilight(
			{
				"menuitem_" + i.id: "#" + self.config["osd_colors"]["menuitem_hilight"],
				"text_" + i.id: "#" + self.config["osd_colors"]["menuitem_hilight_text"],
			},
		)

	def on_event(self, daemon, what, data):
		if self._submenu:
			return self._submenu.on_event(daemon, what, data)
		if what == self._control_with:
			x, y = data
			# Special case, both confirm_with and cancel_with can be set to STICK
			if self._cancel_with == STICK and self._control_with == STICK:
				if self._control_equals_cancel(daemon, x, y):
					return None

			if self.rotation:
				rx = x * cos(self.rotation) - y * sin(self.rotation)
				ry = x * sin(self.rotation) + y * cos(self.rotation)
				x, y = rx, ry

			max_w = self.get_allocation().width * self.scale - (self.cursor.get_allocation().width * 1.0)
			max_h = self.get_allocation().height * self.scale - (self.cursor.get_allocation().height * 1.0)
			cx = ((x * 0.75 / (STICK_PAD_MAX * 2.0)) + 0.5) * max_w
			cy = (0.5 - (y * 0.75 / (STICK_PAD_MAX * 2.0))) * max_h

			cx -= self.cursor.get_allocation().width * 0.5
			cy -= self.cursor.get_allocation().height * 0.5
			self.f.move(self.cursor, int(cx), int(cy))

			if abs(x) + abs(y) > RadialMenu.MIN_DISTANCE:
				angle = atan2(x, y) * 180.0 / PI
				half_width = 180.0 / len(self.items)
				for i in self.items:
					if abs(degdiff(i.a, angle)) < half_width:
						if self._selected != i:
							if self.feedback and self.controller:
								self.controller.feedback(*self.feedback)
							self.select(i)
		else:
			return Menu.on_event(self, daemon, what, data)
		return None


if __name__ == "__main__":
	import gi

	gi.require_version("Gtk", "3.0")
	gi.require_version("Rsvg", "2.0")
	gi.require_version("GdkX11", "3.0")

	from scc.tools import init_logging

	init_logging()

	m = RadialMenu()
	if not m.parse_argumets(sys.argv):
		sys.exit(1)
	m.run()
	if m.get_exit_code() == 0:
		print(m.get_selected_item_id())
	sys.exit(m.get_exit_code())



#!/usr/bin/env python3
"""SC-Controller - On Screen Keyboard Binding Editor

Edits '.scc-osd.keyboard.sccprofile', profile used by on screen keyboard
"""

import logging
import os

from scc.actions import Action
from scc.constants import STICK, SCButtons
from scc.gui.binding_editor import BindingEditor
from scc.gui.controller_widget import STICKS, TRIGGERS
from scc.gui.editor import Editor
from scc.gui.parser import GuiActionParser
from scc.osd.keyboard import Keyboard as OSDKeyboard
from scc.paths import get_profiles_path
from scc.profile import Profile
from scc.tools import _, find_profile

log = logging.getLogger("OSKEdit")


class OSKBindingEditor(Editor, BindingEditor):
	GLADE = "osk_binding_editor.glade"

	def __init__(self, app):
		BindingEditor.__init__(self, app)
		self.app = app
		self.gladepath = app.gladepath
		self.imagepath = app.imagepath
		self.current = Profile(GuiActionParser())
		self.current.load(find_profile(OSDKeyboard.OSK_PROF_NAME))
		self.setup_widgets()

	def setup_widgets(self):
		Editor.setup_widgets(self)
		self.create_binding_buttons(use_icons=False, enable_press=False)

	def show_editor(self, id):
		if id in STICKS:
			ae = self.choose_editor(self.current.stick, _("Stick"))
			ae.set_input(STICK, self.current.stick, mode=Action.AC_OSK)
			ae.show(self.window)
		elif id in SCButtons.__members__.values():
			title = _("%s Button") % (id.name,)
			ae = self.choose_editor(self.current.buttons[id], title)
			ae.set_input(id, self.current.buttons[id], mode=Action.AC_OSK)
			ae.show(self.window)
		elif id in TRIGGERS:
			ae = self.choose_editor(self.current.triggers[id], _("%s Trigger") % (id,))
			ae.set_input(id, self.current.triggers[id], mode=Action.AC_OSK)
			ae.show(self.window)

	def on_action_chosen(self, id, action, mark_changed=True):
		self.set_action(self.current, id, action)
		self.save_profile()

	def save_profile(self, *a):
		"""Saves osk profile from 'profile' object into 'giofile'.
		Calls on_profile_saved when done
		"""
		self.current.save(os.path.join(get_profiles_path(), OSDKeyboard.OSK_PROF_NAME + ".sccprofile"))
		# OSK reloads profile when daemon reports configuration change
		self.app.dm.reconfigure()



"""SC Controller - Controller Registration - Grabs

Helper classes for grabbing buttons and axes from physical gamepads.

"""

import logging

from scc.constants import STICK_PAD_MAX, STICK_PAD_MIN
from scc.gui.creg.constants import X, Y
from scc.gui.creg.data import AxisData, DPadEmuData
from scc.tools import _, nameof

log = logging.getLogger("CReg.grabs")


class InputGrabber:
	"""Base class for input grabbing. Waits for physical button being pressed by default."""

	def __init__(self, parent, what, text=_("Press a button...")):
		self.parent = parent
		self.what = what
		self.set_message(text)
		self.dlgPressButton = parent.builder.get_object("dlgPressButton")
		self.dlgPressButton.show()

	def set_message(self, text):
		self.parent.builder.get_object("lblPressButton").set_text(text)

	def cancel(self):
		self.dlgPressButton.hide()
		self.parent._grabber = None

	def on_button(self, keycode, pressed):
		if not pressed:
			self.set_mapping(keycode, self.what)

	def set_mapping(self, keycode, what):
		parent = self.parent

		if isinstance(what, AxisData) and what in parent._mappings.values():
			for c in parent._mappings.copy():
				if parent._mappings[c] == what:
					del parent._mappings[c]

		parent._mappings[keycode] = what
		log.debug("Reassigned %s to %s", keycode, what)

		if nameof(what) in parent._unassigned:
			parent._unassigned.remove(nameof(what))
			parent.unhilight(nameof(what))

		self.parent.generate_unassigned()
		self.parent.generate_raw_data()
		self.cancel()

	def on_axis(self, number, value):
		pass


class TriggerGrabber(InputGrabber):
	"""InputGrabber modified to grab trigger bindings.
	That may be button or axis with at least 0-250 range is accepted.
	"""

	def __init__(self, parent, what, text=_("Pull a trigger...")):
		InputGrabber.__init__(self, parent, what, text)
		self.orig_pos = {k: parent._input_axes[k] for k in parent._input_axes}
		self.new_pos = {k: parent._input_axes[k] for k in parent._input_axes}

	def on_button(self, keycode, pressed):
		if not pressed:
			self.set_mapping(keycode, self.what)
			self.what.min = STICK_PAD_MIN
			self.what.max = STICK_PAD_MAX

	def on_axis(self, number, value):
		if number > 50:
			# TODO: Remove this condition
			return
		self.new_pos[number] = value
		if number not in self.orig_pos:
			self.orig_pos[number] = 0

		# Get avgerage absolute change for all axes
		avg = float(sum([abs(self.orig_pos[k] - self.new_pos[k]) for k in self.new_pos])) / float(len(self.new_pos))

		# Get absolute change for _this_ axis
		change = abs(self.orig_pos[number] - self.new_pos[number])
		if change > 2 and change > avg * 0.5:
			# TODO: change > 2 may be too strict
			# if there is pad going from -1 to 1 somewhere around
			self.axis_change(number, value, change)

	def axis_change(self, number, value, change):
		if value > 250:
			self.what.reset()
			self.set_mapping(number, self.what)
			self.parent.generate_unassigned()
			self.parent.generate_raw_data()
			self.cancel()


class StickGrabber(TriggerGrabber):
	"""InputGrabber modified to grab stick or pad bindings, in two phases for
	both X and Y axis.
	"""

	def __init__(self, parent, what):
		TriggerGrabber.__init__(self, parent, what, text=_("Move stick left and right..."))
		self.xy = X
		self.grabbed = [None, None]

	def on_button(self, keycode, pressed):
		# if len(self.grabbed) == 2 and self.grabbed[X] != None:
		# # Already grabbed one axis, don't grab buttons
		# return
		if keycode in self.grabbed:
			# Don't allow same button to be used twice
			return
		if not pressed:
			if len(self.grabbed) < 4:
				self.grabbed = [None] * 4
			if self.grabbed[0] is None:
				self.grabbed[0] = keycode
				self.set_message(_("Move DPAD to right"))
			elif self.grabbed[1] is None:
				self.grabbed[1] = keycode
				self.set_message(_("Move DPAD up"))
			elif self.grabbed[2] is None:
				self.grabbed[2] = keycode
				self.set_message(_("Move DPAD down"))
			elif self.grabbed[3] is None:
				self.grabbed[3] = keycode
				self.set_message(str(self.grabbed))
				grabbed = [] + self.grabbed
				for w in self.what:
					for negative in (False, True):
						keycode, grabbed = grabbed[0], grabbed[1:]
						w.reset()
						self.set_mapping(keycode, DPadEmuData(w, negative))
				self.parent.generate_unassigned()
				self.parent.generate_raw_data()
				self.cancel()

	def on_axis(self, number, value):
		if len(self.grabbed) > 2:
			# Already started grabbing 4 buttons, don't grab axes now
			return
		if self.xy == X:
			self.grabbed[X] = number
			self.xy = Y
			self.set_message(_("Move stick up and down..."))
		elif number != self.grabbed[X]:
			self.grabbed[Y] = number
			for i in range(len(self.grabbed)):
				self.what[i].reset()
				self.set_mapping(self.grabbed[i], self.what[i])
			self.parent.generate_unassigned()
			self.parent.generate_raw_data()
			self.cancel()



"""Copied directly from python because I can't find way how to override _iterencode_list, function burried in 7th level of hell.

Only idea here is to have lists encoded in single line.
"""

import re

try:
	from _json import encode_basestring_ascii as c_encode_basestring_ascii
except ImportError:
	c_encode_basestring_ascii = None
try:
	from _json import make_encoder as c_make_encoder
except ImportError:
	c_make_encoder = None

ESCAPE = re.compile(r'[\x00-\x1f\\"\b\f\n\r\t]')
ESCAPE_ASCII = re.compile(r'([\\"]|[^\ -~])')
HAS_UTF8 = re.compile(r"[\x80-\xff]")
ESCAPE_DCT = {
	"\\": "\\\\",
	'"': '\\"',
	"\b": "\\b",
	"\f": "\\f",
	"\n": "\\n",
	"\r": "\\r",
	"\t": "\\t",
}
for i in range(0x20):
	ESCAPE_DCT.setdefault(chr(i), f"\\u{i:04x}")
	# ESCAPE_DCT.setdefault(chr(i), '\\u%04x' % (i,))

INFINITY = float("inf")
FLOAT_REPR = repr


def encode_basestring(s):
	"""Return a JSON representation of a Python string"""

	def replace(match):
		return ESCAPE_DCT[match.group(0)]

	return '"' + ESCAPE.sub(replace, s) + '"'


def py_encode_basestring_ascii(s):
	"""Return an ASCII-only JSON representation of a Python string"""
	if isinstance(s, str) and HAS_UTF8.search(s) is not None:
		s = s

	def replace(match):
		s = match.group(0)
		try:
			return ESCAPE_DCT[s]
		except KeyError:
			n = ord(s)
			if n < 0x10000:
				return f"\\u{n:04x}"
				# return '\\u%04x' % (n,)
			# surrogate pair
			n -= 0x10000
			s1 = 0xD800 | ((n >> 10) & 0x3FF)
			s2 = 0xDC00 | (n & 0x3FF)
			return f"\\u{s1:04x}\\u{s2:04x}"
				# return '\\u%04x\\u%04x' % (s1, s2)

	return '"' + str(ESCAPE_ASCII.sub(replace, s)) + '"'


encode_basestring_ascii = c_encode_basestring_ascii or py_encode_basestring_ascii


class JSONEncoder:
	"""Extensible JSON <http://json.org> encoder for Python data structures.

	Supports the following objects and types by default:

	+-------------------+---------------+
	| Python			| JSON		  |
	+===================+===============+
	| dict			  | object		|
	+-------------------+---------------+
	| list, tuple	   | array		 |
	+-------------------+---------------+
	| str, unicode	  | string		|
	+-------------------+---------------+
	| int, long, float  | number		|
	+-------------------+---------------+
	| True			  | true		  |
	+-------------------+---------------+
	| False			 | false		 |
	+-------------------+---------------+
	| None			  | null		  |
	+-------------------+---------------+

	To extend this to recognize other objects, subclass and implement a
	``.default()`` method with another method that returns a serializable
	object for ``o`` if possible, otherwise it should call the superclass
	implementation (to raise ``TypeError``).

	"""

	item_separator = ", "
	key_separator = ": "

	def __init__(
		self,
		skipkeys=False,
		ensure_ascii=True,
		check_circular=True,
		allow_nan=True,
		sort_keys=False,
		indent=None,
		separators=None,
		encoding="utf-8",
		default=None,
	):
		r"""Constructor for JSONEncoder, with sensible defaults.

		If skipkeys is false, then it is a TypeError to attempt
		encoding of keys that are not str, int, long, float or None.  If
		skipkeys is True, such items are simply skipped.

		If *ensure_ascii* is true (the default), all non-ASCII
		characters in the output are escaped with \uXXXX sequences,
		and the results are str instances consisting of ASCII
		characters only.  If ensure_ascii is False, a result may be a
		unicode instance.  This usually happens if the input contains
		unicode strings or the *encoding* parameter is used.

		If check_circular is true, then lists, dicts, and custom encoded
		objects will be checked for circular references during encoding to
		prevent an infinite recursion (which would cause an OverflowError).
		Otherwise, no such check takes place.

		If allow_nan is true, then NaN, Infinity, and -Infinity will be
		encoded as such.  This behavior is not JSON specification compliant,
		but is consistent with most JavaScript based encoders and decoders.
		Otherwise, it will be a ValueError to encode such floats.

		If sort_keys is true, then the output of dictionaries will be
		sorted by key; this is useful for regression tests to ensure
		that JSON serializations can be compared on a day-to-day basis.

		If indent is a non-negative integer, then JSON array
		elements and object members will be pretty-printed with that
		indent level.  An indent level of 0 will only insert newlines.
		None is the most compact representation.  Since the default
		item separator is ', ',  the output might include trailing
		whitespace when indent is specified.  You can use
		separators=(',', ': ') to avoid this.

		If specified, separators should be a (item_separator, key_separator)
		tuple.  The default is (', ', ': ').  To get the most compact JSON
		representation you should specify (',', ':') to eliminate whitespace.

		If specified, default is a function that gets called for objects
		that can't otherwise be serialized.  It should return a JSON encodable
		version of the object or raise a ``TypeError``.

		If encoding is not None, then all input strings will be
		transformed into unicode using that encoding prior to JSON-encoding.
		The default is UTF-8.

		"""
		self.skipkeys = skipkeys
		self.ensure_ascii = ensure_ascii
		self.check_circular = check_circular
		self.allow_nan = allow_nan
		self.sort_keys = sort_keys
		self.indent = indent
		if separators is not None:
			self.item_separator, self.key_separator = separators
		if default is not None:
			self.default = default
		self.encoding = encoding

	def default(self, o):
		"""Implement this method in a subclass such that it returns
		a serializable object for ``o``, or calls the base implementation
		(to raise a ``TypeError``).

		For example, to support arbitrary iterators, you could
		implement default like this::

			def default(self, o):
				try:
					iterable = iter(o)
				except TypeError:
					pass
				else:
					return list(iterable)
				# Let the base class default method raise the TypeError
				return JSONEncoder.default(self, o)

		"""
		raise TypeError(repr(o) + " is not JSON serializable")

	def encode(self, o):
		"""Return a JSON string representation of a Python data structure.

		>>> JSONEncoder().encode({"foo": ["bar", "baz"]})
		'{"foo": ["bar", "baz"]}'

		"""
		# This is for extremely simple cases and benchmarks.
		if isinstance(o, str):
			_encoding = self.encoding
			if _encoding is not None and not (_encoding == "utf-8"):
				# o = o.decode(_encoding)
				o = o

			if self.ensure_ascii:
				return encode_basestring_ascii(o)
			return encode_basestring(o)
		# This doesn't pass the iterator directly to ''.join() because the
		# exceptions aren't as detailed.  The list call should be roughly
		# equivalent to the PySequence_Fast that ''.join() would do.
		chunks = self.iterencode(o, _one_shot=True)
		if not isinstance(chunks, (list, tuple)):
			chunks = list(chunks)
		return "".join(chunks)

	def iterencode(self, o, _one_shot=False):
		"""Encode the given object and yield each string
		representation as available.

		For example::

			for chunk in JSONEncoder().iterencode(bigobject):
				mysocket.write(chunk)

		"""
		if self.check_circular:
			markers = {}
		else:
			markers = None
		if self.ensure_ascii:
			_encoder = encode_basestring_ascii
		else:
			_encoder = encode_basestring
		if self.encoding != "utf-8":

			def _encoder(o, _orig_encoder=_encoder, _encoding=self.encoding):
				# if isinstance(o, str):
				# o = o.decode(_encoding)
				return _orig_encoder(o)

		def floatstr(o, allow_nan=self.allow_nan, _repr=FLOAT_REPR, _inf=INFINITY, _neginf=-INFINITY):
			# Check for specials.  Note that this type of test is processor
			# and/or platform-specific, so do tests which don't depend on the
			# internals.

			if o != o:
				text = "NaN"
			elif o == _inf:
				text = "Infinity"
			elif o == _neginf:
				text = "-Infinity"
			else:
				return _repr(o)

			if not allow_nan:
				raise ValueError("Out of range float values are not JSON compliant: " + repr(o))

			return text

		if _one_shot and c_make_encoder is not None and self.indent is None and not self.sort_keys:
			_iterencode = c_make_encoder(
				markers,
				self.default,
				_encoder,
				self.indent,
				self.key_separator,
				self.item_separator,
				self.sort_keys,
				self.skipkeys,
				self.allow_nan,
			)
		else:
			_iterencode = _make_iterencode(
				markers,
				self.default,
				_encoder,
				self.indent,
				floatstr,
				self.key_separator,
				self.item_separator,
				self.sort_keys,
				self.skipkeys,
				_one_shot,
			)
		return _iterencode(o, 0)


def _make_iterencode(
	markers,
	_default,
	_encoder,
	_indent,
	_floatstr,
	_key_separator,
	_item_separator,
	_sort_keys,
	_skipkeys,
	_one_shot,
	## HACK: hand-optimized bytecode; turn globals into locals
	ValueError=ValueError,
	basestring=str,
	dict=dict,
	float=float,
	id=id,
	int=int,
	isinstance=isinstance,
	list=list,
	long=int,
	str=str,
	tuple=tuple,
):
	def _iterencode_list(lst, _current_indent_level):
		if not lst:
			yield "[]"
			return
		if markers is not None:
			markerid = id(lst)
			if markerid in markers:
				raise ValueError("Circular reference detected")
			markers[markerid] = lst
		buf = "["
		if False:  #  _indent is not None:
			_current_indent_level += 1
			newline_indent = "\n" + (" " * (_indent * _current_indent_level))
			separator = _item_separator + newline_indent
			buf += newline_indent
		else:
			newline_indent = None
			separator = _item_separator
		newline_indent = None
		separator = _item_separator
		first = True
		for value in lst:
			if first:
				first = False
			else:
				buf = separator
			if isinstance(value, basestring):
				yield buf + _encoder(value)
			elif value is None:
				yield buf + "null"
			elif value is True:
				yield buf + "true"
			elif value is False:
				yield buf + "false"
			elif isinstance(value, (int, long)):
				yield buf + str(value)
			elif isinstance(value, float):
				yield buf + _floatstr(value)
			else:
				yield buf
				if isinstance(value, (list, tuple)):
					chunks = _iterencode_list(value, _current_indent_level)
				elif isinstance(value, dict):
					chunks = _iterencode_dict(value, _current_indent_level)
				else:
					chunks = _iterencode(value, _current_indent_level)
				for chunk in chunks:
					yield chunk
		if newline_indent is not None:
			_current_indent_level -= 1
			yield "\n" + (" " * (_indent * _current_indent_level))
		yield "]"
		if markers is not None:
			del markers[markerid]

	def _iterencode_dict(dct, _current_indent_level):
		if not dct:
			yield "{}"
			return
		if markers is not None:
			markerid = id(dct)
			if markerid in markers:
				raise ValueError("Circular reference detected")
			markers[markerid] = dct
		yield "{"
		if _indent is not None:
			_current_indent_level += 1
			newline_indent = "\n" + (" " * (_indent * _current_indent_level))
			item_separator = _item_separator + newline_indent
			yield newline_indent
		else:
			newline_indent = None
			item_separator = _item_separator
		first = True
		if _sort_keys:
			items = sorted(dct.items(), key=lambda kv: kv[0])
		else:
			items = dct.iteritems()
		for key, value in items:
			if isinstance(key, basestring):
				pass
			# JavaScript is weakly typed for these, so it makes sense to
			# also allow them.  Many encoders seem to do something like this.
			elif isinstance(key, float):
				key = _floatstr(key)
			elif key is True:
				key = "true"
			elif key is False:
				key = "false"
			elif key is None:
				key = "null"
			elif isinstance(key, (int, long)):
				key = str(key)
			elif _skipkeys:
				continue
			else:
				raise TypeError("key " + repr(key) + " is not a string")
			if first:
				first = False
			else:
				yield item_separator
			yield _encoder(key)
			yield _key_separator
			if isinstance(value, basestring):
				yield _encoder(value)
			elif value is None:
				yield "null"
			elif value is True:
				yield "true"
			elif value is False:
				yield "false"
			elif isinstance(value, (int, long)):
				yield str(value)
			elif isinstance(value, float):
				yield _floatstr(value)
			else:
				if isinstance(value, (list, tuple)):
					chunks = _iterencode_list(value, _current_indent_level)
				elif isinstance(value, dict):
					chunks = _iterencode_dict(value, _current_indent_level)
				else:
					chunks = _iterencode(value, _current_indent_level)
				for chunk in chunks:
					yield chunk
		if newline_indent is not None:
			_current_indent_level -= 1
			yield "\n" + (" " * (_indent * _current_indent_level))
		yield "}"
		if markers is not None:
			del markers[markerid]

	def _iterencode(o, _current_indent_level):
		if isinstance(o, basestring):
			yield _encoder(o)
		elif o is None:
			yield "null"
		elif o is True:
			yield "true"
		elif o is False:
			yield "false"
		elif isinstance(o, (int, long)):
			yield str(o)
		elif isinstance(o, float):
			yield _floatstr(o)
		elif isinstance(o, (list, tuple)):
			for chunk in _iterencode_list(o, _current_indent_level):
				yield chunk
		elif isinstance(o, dict):
			for chunk in _iterencode_dict(o, _current_indent_level):
				yield chunk
		else:
			if markers is not None:
				markerid = id(o)
				if markerid in markers:
					raise ValueError("Circular reference detected")
				markers[markerid] = o
			o = _default(o)
			for chunk in _iterencode(o, _current_indent_level):
				yield chunk
			if markers is not None:
				del markers[markerid]

	return _iterencode



#!/usr/bin/env python3
"""SC-Controller - Controller Registration - Tester

Class that interacts with `scc hid_test` and `scc evdev_test` commands.
"""

import logging

from gi.repository import Gio, GObject

from scc.tools import find_binary

log = logging.getLogger("CReg.Tester")


class Tester(GObject.GObject):
	"""List of signals:
	ready ()
		Emited when subprocess signalizes it's ready to send data
	error (code)
		Emited when driver test subprocess exits with non-zero return code
	finished ()
		Emited when driver test subprocess exits with zero return code
	axis (number, value)
		Emited when position on axis is changed
	button (keycode, pressed)
		Emited when button on tested gamepad is pressed or released
	"""

	__gsignals__ = {
		"error": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
		"ready": (GObject.SignalFlags.RUN_FIRST, None, ()),
		"finished": (GObject.SignalFlags.RUN_FIRST, None, ()),
		"axis": (GObject.SignalFlags.RUN_FIRST, None, (int, int)),
		"button": (GObject.SignalFlags.RUN_FIRST, None, (int, bool)),
	}

	def __init__(self, driver, device_id):
		GObject.GObject.__init__(self)
		self.buffer = b""
		self.buttons = []
		self.axes = []
		self.subprocess = None
		self.driver = driver
		self.device_id = device_id
		self.errorred = False  # To prevent sending 'error' signal multiple times

	def __del__(self):
		if self.subprocess:
			self.subprocess.send_signal(9)

	def start(self):
		"""Starts driver test subprocess"""
		cmd = [find_binary("scc")] + ["test_" + self.driver, self.device_id]
		self.subprocess = Gio.Subprocess.new(cmd, Gio.SubprocessFlags.STDOUT_PIPE)
		self.subprocess.wait_async(None, self._on_finished)
		self.subprocess.get_stdout_pipe().read_bytes_async(32, 0, None, self._on_read)

	def stop(self):
		if self.subprocess:
			self.subprocess.send_signal(2)  # Sigint

	def _on_finished(self, subprocess, result):
		subprocess.wait_finish(result)
		if self.errorred:
			return
		if subprocess.get_exit_status() == 0:
			self.emit("finished")
		else:
			self.errorred = True
			self.emit("error", subprocess.get_exit_status())

	def _on_read(self, stream, result):
		try:
			data = stream.read_bytes_finish(result).get_data()
		except Exception as e:
			log.exception(e)
			self.subprocess.send_signal(2)
			if not self.errorred:
				self.errorred = True
				self.emit("error", 1)
			return
		if len(data) > 0:
			self.buffer += data
			while b"\n" in self.buffer:
				line, self.buffer = self.buffer.split(b"\n", 1)
				try:
					self._on_line(line)
				except Exception as e:
					log.exception(e)
			self.subprocess.get_stdout_pipe().read_bytes_async(32, 0, None, self._on_read)

	def _on_line(self, line):
		if line.startswith(b"Axis"):
			trash, number, value = line.split(b" ")
			number, value = int(number), int(value)
			self.emit("axis", number, value)
		elif line.startswith(b"ButtonPress"):
			trash, code = line.split(b" ")
			self.emit("button", int(code), True)
		elif line.startswith(b"ButtonRelease"):
			trash, code = line.split(b" ")
			self.emit("button", int(code), False)
		elif line.startswith(b"Ready"):
			self.emit("ready")
		elif line.startswith(b"Axes:"):
			self.axes = [int(x) for x in line.split(b" ")[1:] if len(x.strip())]
		elif line.startswith(b"Buttons:"):
			self.buttons = [int(x) for x in line.split(b" ")[1:] if len(x.strip())]
